"""
Validate and clean request parameters for our endpoints using Django forms
"""
from datetime import MAXYEAR
from functools import reduce

from django import forms
from django.conf import settings

import arrow

from .defaults import RELATED_CONTENT_BOOST
from .filter_definitions import FILTERS, AvailabilityFilterDefinition

# Instantiate filter fields for each filter defined in settings
# It is of the form:
#     {
#         subjects: (
#             ArrayField(required=False, base_type=forms.CharField(max_length=50)),
#             True,
#         ),
#         subjects_include: (
#             CharField(required=False, max_length=20),
#             False,
#         ),
#         languages: (
#             MultipleChoiceField(required=False, choices=LANGUAGES),
#             True,
#         ),
#     }
# Where the first item is an instance of the form field and the second is a boolean
# indicating whether this formfield expects a list or a single value.
FILTER_FIELDS = {
    key: value
    for filter_definition in FILTERS.values()
    for key, value in filter_definition.get_form_fields().items()
}


class SearchForm(forms.Form):
    """Validate the query string params in a search request."""

    OBJECTS, FILTERS = "objects", "filters"

    SCOPE_CHOICES = ((OBJECTS, "Objects"), (FILTERS, "Filters"))

    limit = forms.IntegerField(required=False, min_value=1, initial=10)
    query = forms.CharField(required=False, min_length=3, max_length=100)
    offset = forms.IntegerField(required=False, min_value=0, initial=0)
    scope = forms.ChoiceField(required=False, choices=SCOPE_CHOICES)


class CourseSearchForm(SearchForm):
    """
    Validate the query string params in the course search request, connect them to filter
    definitions and generate Elasticsearch queries.
    """

    def __init__(self, *args, data=None, **kwargs):
        """
        Adapt the search form to handle filters:
        - Fix the QueryDict value getter to properly handle multi-value parameters,
        - Add a field instance to the form for each filter,
        - Define the `states` property as it is used by several methods.
        """
        # QueryDict/MultiValueDict breaks lists: we need to fix it
        data_fixed = (
            {
                k: data.getlist(k)
                # Form fields are marked to expect lists as input or not as explained above
                if (k in FILTER_FIELDS and FILTER_FIELDS[k][1] is True) else v[0]
                for k, v in data.lists()
            }
            if data
            else {}
        )

        super().__init__(data=data_fixed, *args, **kwargs)
        self.fields.update({k: v[0] for k, v in FILTER_FIELDS.items()})
        self.states = None

    def clean_availability(self):
        """
        Calculate and set the list of states relevant with the current availability filter.
        e.g. if we filter on OPEN courses, only the course runs in state 0 (ongoing open) or
        1 (future open) should be considered for sorting and computation of the course's state.
        """
        availabilities = self.cleaned_data.get("availability", [])
        if AvailabilityFilterDefinition.OPEN in availabilities:
            self.states = [0, 1]
        elif AvailabilityFilterDefinition.ONGOING in availabilities:
            self.states = [0, 4]
        elif AvailabilityFilterDefinition.COMING_SOON in availabilities:
            self.states = [1, 2, 3]
        elif AvailabilityFilterDefinition.ARCHIVED in availabilities:
            self.states = [5]

        return availabilities

    def get_script_fields(self):
        """
        Build the part of the Elasticseach query that defines script fields ie fields that can not
        be indexed because they are dynamic and shoud be calculated at query time:
        - ms_since_epoch: evolves with time to stay relevant on course dates even if the ES
          instance and/or the Django server are long running.
        - languages and states: depend on the filters applied by the user so that we only take into
          consideration course runs that are interesting for this search.
        - use_case: the script is used both for sorting and field computations because most of the
          code is common and their is no other way to share code.

        Note: we use script storage to save time on the script compilation, which is an expensive
        operation. We'll only do it once at bootstrap time.
        """
        return {
            "state": {
                "script": {
                    "id": "state",
                    "params": {
                        "languages": self.cleaned_data.get("languages") or None,
                        "ms_since_epoch": arrow.utcnow().timestamp * 1000,
                        "states": self.states,
                        "use_case": "field",
                    },
                }
            }
        }

    def get_sorting_script(self):
        """
        Build the part of the Elasticseach query that defines sorting. We use a script for sorting
        because we sort courses based on the complex and dynamic status of their course runs which
        are under a nested field. The parameters passed to the script are up-to-date at query time:
        - ms_since_epoch: evolves with time to stay relevant on course dates even if the ES
          instance and/or the Django server are long running,
        - languages and states: depend on the filters applied by the user so that we only take into
          consideration course runs that are interesting for this search,
        - max_date: passed as parameter to optimize script compilation,
        - use_case: the script is used both for sorting and field computations because most of the
          code is common and their is no other way to share code.


        Call the relevant sorting script for courses lists, regenerating the parameters on each
        call. This will allow the ms_since_epoch value to stay relevant even if the ES instance
        and/or the Django server are long running.

        The list of languages and states are passed to the script because the context of the
        search defines which course runs are relevant or not for sorting.

        Note: we use script storage to save time on the script compilation, which is an expensive
        operation. We'll only do it once at bootstrap time.
        """
        return {
            "_script": {
                "order": "asc",
                "script": {
                    "id": "state",
                    "params": {
                        "languages": self.cleaned_data.get("languages") or None,
                        "max_date": arrow.get(MAXYEAR, 12, 31).timestamp * 1000,
                        "ms_since_epoch": arrow.utcnow().timestamp * 1000,
                        "states": self.states,
                        "use_case": "sorting",
                    },
                },
                "type": "number",
            }
        }

    def get_queries(self):
        """
        Aggregate queries from each filter definition.
        The full text query is treated directly as it does not require heavy lifting.

        Returns:
        --------
            List[Dict]: a list of dictionaries each mapping the name of a filter with a query
                fragment that composes the global Elasticsearch query. For example:
                [
                    {
                        "key": "new",
                        "fragment": [{'term': {'is_new': True}}]
                    }
                    ...
                ]
        """
        queries = []

        # Add the query fragments of each filter definition to the list of queries
        for filter_definition in FILTERS.values():
            queries = queries + filter_definition.get_query_fragment(self.cleaned_data)

        # Full text search is a regular (multilingual) match query
        full_text = self.cleaned_data.get("query")
        if full_text:
            queries.append(
                {
                    "key": "query",
                    "fragment": [
                        {
                            "bool": {
                                "should": [
                                    {
                                        "multi_match": {
                                            "fields": ["description.*", "title.*"],
                                            "query": full_text,
                                            "type": "cross_fields",
                                        }
                                    },
                                    {
                                        "multi_match": {
                                            "boost": getattr(
                                                settings,
                                                "RICHIE_RELATED_CONTENT_BOOST",
                                                RELATED_CONTENT_BOOST,
                                            ),
                                            "fields": [
                                                "categories_names.*",
                                                "organizations_names.*",
                                            ],
                                            "query": full_text,
                                            "type": "cross_fields",
                                        }
                                    },
                                    {
                                        "multi_match": {
                                            "boost": getattr(
                                                settings,
                                                "RICHIE_RELATED_CONTENT_BOOST",
                                                RELATED_CONTENT_BOOST,
                                            ),
                                            "fields": ["persons_names.*"],
                                            "query": full_text,
                                            "type": "cross_fields",
                                        }
                                    },
                                ]
                            }
                        }
                    ],
                }
            )

        return queries

    def build_es_query(self):
        """
        Build the actual Elasticsearch search query and aggregation query from the fragments
        returned by each filter definition.

        Returns:
        --------
            Tuple:
            - limit (int): the maximum number of results to be returned by Elasticsearch,
            - offset (int): the offset from which results are returned (for pagination),
            - query (Dict): the raw Elasticsearch query as per:
              https://elastic.co/guide/en/elasticsearch/reference/current/search.html
            - aggs (Dict): the raw Elasticsearch aggregation as per:
              https://elastic.co/guide/en/elasticsearch/reference/current/search-aggregations.html

        """
        # queries is an array of individual queries that will be combined through "bool" before
        # we pass them to ES. See the docs on bool queries.
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-bool-query.html
        queries = self.get_queries()

        # Default to a match_all query
        if not queries:
            query = {"match_all": {}}
        else:
            # Concatenate all the sub-queries lists together to form the queries list
            query = {
                "bool": {
                    "must":
                    # queries => map(pluck("fragment")) => flatten()
                    [clause for kf_pair in queries for clause in kf_pair["fragment"]]
                }
            }

        # Concatenate our hardcoded filters query fragments with organizations and categories
        # terms aggregations build on-the-fly
        aggs = {
            "all_courses": {
                "global": {},
                "aggregations": reduce(
                    # Merge all the partial aggregations dicts together
                    lambda acc, aggs_fragment: {**acc, **aggs_fragment},
                    # Generate a partial aggregations dict (an aggs_fragment) for each filter
                    [
                        filter.get_aggs_fragment(queries, self.cleaned_data)
                        for filter in FILTERS.values()
                    ],
                    {},
                ),
            }
        }

        return (
            self.cleaned_data.get("limit"),
            self.cleaned_data.get("offset") or 0,
            query,
            aggs,
        )


class ItemSearchForm(SearchForm):
    """Generate Elasticsearch queries for the category/organization indices."""

    def __init__(self, *args, data=None, **kwargs):
        """Fix the QueryDict value getter to properly handle multi-value parameters."""
        # QueryDict/MultiValueDict breaks lists: we need to fix it
        data_fixed = {k: v[0] for k, v in data.lists()} if data else {}
        super().__init__(data=data_fixed, *args, **kwargs)

    def build_es_query(self, kind=None):
        """
        Build the actual Elasticsearch search query for category/organization indices.

        Returns:
        --------
            Tuple:
            - limit (int): the maximum number of results to be returned by Elasticsearch,
            - offset (int): the offset from which results are returned (for pagination),
            - query (Dict): the raw Elasticsearch query as per:
              https://elastic.co/guide/en/elasticsearch/reference/current/search.html

        """
        # Create an array to put the clauses of our bool/must query
        clauses = []

        # Add a term filter by kind for the objects that need it (eg. `kind="subjects"`)
        if kind:
            clauses.append({"term": {"kind": kind}})

        # Add a match on the name field if it was handed by the client
        full_text = self.cleaned_data.get("query")
        if full_text:
            clauses.append({"match": {"title.*": {"query": full_text}}})

        # Build the query around the clauses if there are any
        if clauses:
            query = {"query": {"bool": {"must": clauses}}}
        # Build a match_all query by default
        else:
            query = {"query": {"match_all": {}}}

        return (
            self.cleaned_data.get("limit"),
            self.cleaned_data.get("offset") or 0,
            query,
        )



"""Custom template tags for the courses application of Richie."""
from django import template
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string

from classytags.arguments import Argument, MultiValueArgument
from classytags.core import Options, Tag
from classytags.utils import flatten_context
from cms.exceptions import PlaceholderNotFound
from cms.templatetags.cms_tags import (
    Placeholder,
    PlaceholderOptions,
    _get_page_by_untyped_arg,
)
from cms.toolbar.utils import get_toolbar_from_request
from cms.utils import get_site_id
from cms.utils.placeholder import validate_placeholder_name
from cms.utils.plugins import get_plugins

# pylint: disable=invalid-name
register = template.Library()


@register.tag("page_placeholder")
class PagePlaceholder(Placeholder):
    """
    This template node is used to output page content and
    is also used in the admin to dynamically generate input fields.
    eg:
    {% page_placeholder "sidebar" page_lookup %}
    {% page_placeholder "sidebar" page_lookup inherit %}
    {% page_placeholder "sidebar" page_lookup as varname %}
    {% page_placeholder "sidebar" page_lookup or %}
        <div>No content</div>
    {% endpage_placeholder %}
    Keyword arguments:
        name: the name of the placeholder
        page_lookup: lookup argument for Page. See `_get_page_by_untyped_arg()` for detailed
            information on the allowed types and their interpretation for the `page_lookup`
            argument.
        inherit : optional argument which if given will result in inheriting
            the content of the placeholder with the same name on parent pages
        varname: context variable name. Output will be added to template context as this variable
            instead of being returned.
        or: optional argument which if given will make the template tag a block
            tag whose content is shown if the placeholder is empty
    """

    name = "page_placeholder"
    options = PlaceholderOptions(
        Argument("name", resolve=False),
        Argument("page_lookup"),
        MultiValueArgument("extra_bits", required=False, resolve=False),
        blocks=[("endpage_placeholder", "nodelist")],
    )

    # pylint: disable=arguments-differ,too-many-arguments
    def render_tag(self, context, name, page_lookup, extra_bits, nodelist=None):
        validate_placeholder_name(name)

        request = context.get("request")

        if request:
            page = _get_page_by_untyped_arg(page_lookup, request, get_site_id(None))

            toolbar = get_toolbar_from_request(request)
            renderer = toolbar.get_content_renderer()

            inherit = "inherit" in extra_bits

            # A placeholder is only editable on its own page
            editable = page == request.current_page

            try:
                content = renderer.render_page_placeholder(
                    slot=name,
                    context=context,
                    inherit=inherit,
                    page=page,
                    nodelist=nodelist,
                    editable=editable,
                )
            except PlaceholderNotFound:
                content = ""
        else:
            content = ""

        if not content and nodelist:
            content = nodelist.render(context)

        if "as" in extra_bits:
            try:
                varname = extra_bits[extra_bits.index("as") + 1]
            except IndexError:
                raise template.TemplateSyntaxError(
                    'the "as" word should be followed by the variable name'
                )
            context[varname] = content
            return ""

        return content


@register.tag("get_placeholder_plugins")
class GetPlaceholderPlugins(Placeholder):
    """
    A template tag that declares a placeholder and sets its plugins as a context variable
    instead of rendering them eg:

        {% get_placeholder_plugins "logo" as varname %}
        {% get_placeholder_plugins "logo" page_lookup as varname %}
        {% get_placeholder_plugins "logo" page_lookup as varname or %}
            <div>No content</div>
        {% endget_placeholder_plugins %}

    This tag can typically be used in association with the block_plugin tag, in a placeholder
    limited to one plugin, to customize the way it is rendered eg:

        {% get_placeholder_plugins "logo" page_lookup as plugins %}
        {% blockplugin plugins.0 %}
            <img src="{% thumbnail instance.picture 300x150 %}"/>
        {% endblockplugin %}

    Keyword arguments:
        name: the name of the placeholder
        page_lookup[optional]: lookup argument for Page. See `_get_page_by_untyped_arg()`
            for detailed information on the allowed types and their interpretation for the
            `page_lookup` argument.
        varname: context variable name. Output will be added to template context as this variable
            instead of being returned.
        or: optional argument which if given will make the template tag a block
            tag whose content is shown if the placeholder is empty

    Note: We must derive from the Placeholder class so that the tag is recognized as a
          placeholder and shown in the structure toolbar.
    """

    name = "get_placeholder_plugins"
    options = PlaceholderOptions(
        Argument("name", resolve=False),
        Argument("page_lookup", required=False),
        "as",
        Argument("varname", resolve=False),
        MultiValueArgument("extra_bits", required=False, resolve=False),
        blocks=[("endget_placeholder_plugins", "nodelist")],
    )

    # pylint: disable=arguments-differ,too-many-arguments
    def render_tag(
        self, context, name, page_lookup, varname, extra_bits, nodelist=None
    ):
        """
        Retrieves the placeholder's plugins and set them as a variable in the template context.
        If the placeholder is empty, render the block as fallback content and return the
        resulting HTML.
        If the placholder is editable, the edit script and markup are added to the HTML content.
        """
        content = ""
        request = context.get("request")

        if request:

            page = _get_page_by_untyped_arg(page_lookup, request, get_site_id(None))

            try:
                placeholder = page.placeholders.get(slot=name)
            except ObjectDoesNotExist:
                context[varname] = []
                return ""
            else:
                context[varname] = [
                    cms_plugin.get_plugin_instance()[0]
                    for cms_plugin in get_plugins(
                        request, placeholder, template=page.get_template()
                    )
                ]

            # Default content if there is no plugins in the placeholder
            if not context[varname] and nodelist:
                content = nodelist.render(context)

            # Add the edit script and markup to the content, only if the placeholder is editable
            # and the visited page is the one on which the placeholder is declared.
            toolbar = get_toolbar_from_request(request)
            if placeholder.page == request.current_page and toolbar.edit_mode_active:
                renderer = toolbar.get_content_renderer()
                data = renderer.get_editable_placeholder_context(placeholder, page=page)
                data["content"] = content
                content = renderer.placeholder_edit_template.format(**data)

        return content


@register.tag()
class BlockPlugin(Tag):
    """
    Like DjangoCMS 'render_plugin_block' but only includes the edit script and markup when
    the related placeholder is editable.

    This issue was raised to DjangoCMS and we need our own template tag until they find a way
    to fix it in DjangoCMS (https://github.com/divio/django-cms/issues/6683).
    """

    name = "blockplugin"
    template = "cms/toolbar/plugin.html"
    options = Options(Argument("plugin"), blocks=[("endblockplugin", "nodelist")])

    # pylint: disable=arguments-differ
    def render_tag(self, context, plugin, nodelist):
        """
        Renders the block for the plugin and returns the resulting HTML leaving the temmpate
        context untouched.
        If the placholder is editable, the edit script and markup are added to the rendered HTML.
        """
        request = context.get("request")
        if not plugin or not request:
            return ""

        # Add the plugin and its rendered content to an internal context
        internal_context = flatten_context(context)
        internal_context["instance"] = plugin
        internal_context["content"] = nodelist.render(context.new(internal_context))

        # Add the edit script and markup to the content, only if the placeholder is editable
        # and the visited page is the one on which the plugin's placeholder is declared.
        toolbar = get_toolbar_from_request(request)
        if plugin.placeholder.page == request.current_page and toolbar.edit_mode_active:
            return render_to_string(self.template, internal_context)

        return internal_context["content"]


@register.filter()
def is_empty_placeholder(page, slot):
    """A template filter to determine if a placeholder is empty.

    This is useful when we don't want to include any wrapper markup in our template unless
    the placeholder unless it actually contains plugins.
    """
    placeholder = page.placeholders.get(slot=slot)
    return not placeholder.cmsplugin_set.exists()



"""Define mixins to easily compose custom FilterDefinition classes."""
import re

from ..defaults import FACET_COUNTS_DEFAULT_LIMIT, FACET_COUNTS_MAX_LIMIT


class TermsQueryMixin:
    """A mixin for filter definitions that need to apply term queries."""

    def get_query_fragment(self, data):
        """Build the query fragments as term queries for each selected value."""
        value_list = data.get(self.name)

        # For terms filters, as the name implies, it's a simple terms fragment
        return (
            [{"key": self.name, "fragment": [{"terms": {self.term: value_list}}]}]
            if value_list
            else []
        )


class TermsAggsMixin:
    """A mixin for filter definitions that need to apply aggregations as term queries."""

    @property
    def aggs_include(self):
        """
        Return a regex to limit the terms on which the field will be faceted. We return ".*" by
        default (no limitation) but it can be overriden in children classes.
        """
        return ".*"

    # pylint: disable=unused-argument
    def get_aggs_fragment(self, queries, data, *args, **kwargs):
        """
        Build the aggregations as a term query that counts all the different values assigned
        to the field.
        """
        # Look for an include parameter in the form data and default to the regex returned
        # by the `aggs_include` property:
        include = data[f"{self.name:s}_include"] or self.aggs_include

        # Use all the query fragments from the queries *but* the one(s) that filter on the
        # current filter
        filter_fragments = [
            clause
            for kf_pair in queries
            for clause in kf_pair["fragment"]
            if kf_pair["key"] is not self.name
        ]

        # Terms aggregation will return the n top facet counts among all the values taken
        # by this field
        terms_aggs = {
            self.name: {
                # Rely on the built-in "terms" aggregation to get everything we need
                "aggregations": {
                    self.name: {
                        "terms": {
                            "field": self.term,
                            "include": include,
                            "min_doc_count": self.min_doc_count,
                            # Use the default limit unless the client is requesting a specific
                            # bunch of docs, in which case we allow for more facet counts
                            "size": FACET_COUNTS_MAX_LIMIT
                            if data[f"{self.name:s}_include"]
                            else FACET_COUNTS_DEFAULT_LIMIT,
                        }
                    }
                },
                # Use all the query fragments from the queries *but* the one(s) that
                # filter on the current filter, as it is handled by ElasticSearch for us
                "filter": {"bool": {"must": filter_fragments}},
            }
        }

        # Filters aggregation for values that were selected in the querystring (we must force
        # them because they may not be in the n top facet counts but we must make sure we keep
        # them so that they remain available as options that the user can see and unselect)
        terms_aggs.update(
            {
                # Create a custom aggregation for each value selected in the querystring
                # and matching the include regex
                # eg `organizations@P-0001` & `organizations@P-0002`
                "{:s}@{:s}".format(self.name, value): {
                    "filter": {
                        "bool": {
                            # Use all the query fragments from the queries *but* the one(s) that
                            # filter on the current filter: we manually add back the only one that
                            # is relevant to the current choice.
                            "must": [{"term": {self.term: value}}]
                            + filter_fragments
                        }
                    }
                }
                for value in data.get(self.name, [])
                # The Elasticsearch include regex matches exact values so we must do the same
                # by adding ^ (resp. $) at the beginning (resp. at the end) of the pattern.
                if re.match(f"^{include:s}$", value)
            }
        )

        return terms_aggs


class ChoicesQueryMixin:
    """A mixin for filter definitions that need to apply predefined queries."""

    def get_query_fragment(self, data):
        """Pick the hardcoded query fragment for each selected value."""
        fragment_map = self.get_fragment_map()
        return [
            {"key": self.name, "fragment": fragment_map[value]}
            for value in data.get(self.name, [])
        ]


class ChoicesAggsMixin:
    """A mixin for filter definitions that need to apply aggregations for predefined choices."""

    # pylint: disable=unused-argument
    def get_aggs_fragment(self, queries, *args, **kwargs):
        """
        Build the aggregations as a set of filters, one for each possible value of the field.
        """
        return {
            # Create a custom aggregation for each possible choice for this filter
            # eg `availability@coming_soon` & `availability@current` & `availability@open`
            "{:s}@{:s}".format(self.name, choice_key): {
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries *but* the one(s) that
                        # filter on the current filter: we manually add back the only one that
                        # is relevant to the current choice.
                        "must": choice_fragment
                        + [
                            clause
                            for kf_pair in queries
                            for clause in kf_pair["fragment"]
                            if kf_pair["key"] is not self.name
                        ]
                    }
                }
            }
            for choice_key, choice_fragment in self.get_fragment_map().items()
        }


class NestedChoicesAggsMixin:
    """
    A mixin for filter definitions that are related to a nested field. The aggregation filter can
    only be recomputed at the level of the parent because it should group all queries of fields
    nested below the parent.
    """

    # pylint: disable=unused-argument
    def get_aggs_fragment(self, queries, data, parent, *args, **kwargs):
        """
        Computing aggregations for a nested field is DIFFICULT because query fragments related to
        nested fields are grouped under their common path. For example combined filters on
        availability and languages would lead to a query like:
            {
                "query": {
                    "nested": {
                        "path": "course_runs",
                        "query": {
                            "bool": {
                                "must": [
                                    {"range": {"course_runs.end": {"lte": "01-01-2019"}}},
                                    {"terms": {"course_runs.languages": ["de", "en", fr"]}},
                                ]
                            }
                        },
                    }
                }
            }
        In this example, computing the facet count for the French filter, is done with the
        following filter (excluding the filter on English and German so we only count French):
            {
                "query": {
                    "nested": {
                        "path": "course_runs",
                        "query": {
                            "bool": {
                                "must": [
                                    {"range": {"course_runs.end": {"lte": "01-01-2019"}}},
                                    {"terms": {"course_runs.languages": ["fr"]}},
                                ]
                            }
                        },
                    }
                }
            }

        This can only be built by calling the parent NestingWrapper with customized filter data.
        """
        return {
            # Create a custom aggregation for each possible choice for this filter
            # eg `availability@coming_soon` & `availability@current` & `availability@open`
            "{:s}@{:s}".format(self.name, choice_key): {
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries (the nesting parent is
                        # responsible for excluding the queries related to nested fields so we
                        # have to manually add them, making sure to apply on the current field
                        # only the current choice.
                        "must": [
                            clause
                            for kf_pair in (
                                queries
                                + parent.get_query_fragment(
                                    # override data with only the current choice
                                    {**data, self.name: [choice_key]}
                                )
                            )
                            for clause in kf_pair["fragment"]
                        ]
                    }
                }
            }
            for choice_key, choice_fragment in self.get_fragment_map().items()
        }



# Generated by Django 2.1.9 on 2019-06-12 15:43

from django.db import migrations

import richie.apps.core.fields.duration
import richie.apps.core.fields.effort


class Migration(migrations.Migration):

    dependencies = [("courses", "0002_auto_20190429_0901")]

    operations = [
        migrations.AddField(
            model_name="course",
            name="effort",
            field=richie.apps.core.fields.effort.EffortField(
                blank=True,
                default_effort_unit="week",
                default_reference_unit="month",
                time_units={
                    "day": ("day", "days"),
                    "hour": ("hour", "hours"),
                    "minute": ("minute", "minutes"),
                    "month": ("month", "months"),
                    "week": ("week", "weeks"),
                },
                max_length=80,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="course",
            name="duration",
            field=richie.apps.core.fields.duration.CompositeDurationField(
                blank=True,
                default_unit="hour",
                time_units={
                    "day": ("day", "days"),
                    "hour": ("hour", "hours"),
                    "minute": ("minute", "minutes"),
                    "month": ("month", "months"),
                    "week": ("week", "weeks"),
                },
                max_length=80,
                null=True,
            ),
        ),
    ]



"""Custom template tags for the search application of Richie."""
from django import template
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import to_locale

from ..defaults import REACT_LOCALES

# pylint: disable=invalid-name
register = template.Library()


@register.filter
def react_locale(value):
    """
    Converts a language (simple ISO639-1 or full language with regions) to an ISO15897 locale
    that is supported by the React frontend.
    """
    value_locale = to_locale(value)
    # pylint: disable=unsupported-membership-test
    if value_locale in REACT_LOCALES:
        return value_locale
    # pylint: disable=not-an-iterable
    for locale in REACT_LOCALES:
        if locale[:2] == value_locale:
            return locale
    raise ImproperlyConfigured(
        f"{value:s} does not correspond to any locale supported by the React frontend."
    )



"""
Declare and configure the models for the courses application
"""
from collections.abc import Mapping
from datetime import MAXYEAR, datetime

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import timezone, translation
from django.utils.functional import lazy
from django.utils.translation import ugettext_lazy as _

import pytz
from cms.extensions.extension_pool import extension_pool
from cms.models import Page, PagePermission
from cms.models.pluginmodel import CMSPlugin
from filer.fields.image import FilerImageField
from filer.models import FolderPermission

from ...core.defaults import ALL_LANGUAGES
from ...core.fields.duration import CompositeDurationField
from ...core.fields.effort import EffortField
from ...core.fields.multiselect import MultiSelectField
from ...core.helpers import get_permissions
from ...core.models import BasePageExtension, PagePluginMixin
from .. import defaults
from .category import Category
from .organization import Organization
from .person import Person
from .role import PageRole

MAX_DATE = datetime(MAXYEAR, 12, 31, tzinfo=pytz.utc)


class CourseState(Mapping):
    """An immutable object to describe a course (resp. course run) state."""

    STATE_CALLS_TO_ACTION = (
        _("enroll now"),
        _("enroll now"),
        _("see details"),
        None,
        None,
        None,
        None,
    )

    STATE_TEXTS = (
        _("closing on"),
        _("starting on"),
        _("starting on"),
        _("enrollment closed"),
        _("on-going"),
        _("archived"),
        _("to be scheduled"),
    )

    def __init__(self, priority, date_time=None):
        """
        Initialize a course state with a priority and optionally a datetime.

        Several states are possible for a course run each of which is given a priority. The
        lower the priority, the more interesting the course run is (a course run open for
        enrollment is more interesting than an archived course run):
          0: a run is on-going and open for enrollment > "closing on": {enrollment_end}
          1: a run is future and open for enrollment > "starting on": {start}
          2: a run is future and not yet open for enrollment > "starting on": {start}
          3: a run is future and no more open for enrollment > "closed": {None}
          4: a run is on-going but closed for enrollment > "on going": {None}
          5: there's a finished run in the past > "archived": {None}
          6: theres's no run with a start date or no run at all > "to be scheduled": {None}
        """
        # Check that `date_time` is set when it should be
        if date_time is None and priority in [0, 1, 2]:
            raise ValidationError(
                "date_time should not be null for a {:d} course state.".format(priority)
            )

        # A special case of being open is when enrollment never ends
        text = self.STATE_TEXTS[priority]
        if priority == 0 and date_time.year == MAXYEAR:
            text = _("forever open")
            date_time = None
        kwargs = {
            "priority": priority,
            "datetime": date_time,
            "call_to_action": self.STATE_CALLS_TO_ACTION[priority],
            "text": text,
        }
        self._d = dict(**kwargs)

    def __iter__(self):
        """Iterate on the inner dictionary."""
        return iter(self._d)

    def __len__(self):
        """Return length of the inner dictionary."""
        return len(self._d)

    def __getitem__(self, key):
        """Access the inner dictionary."""
        return self._d[key]

    def __lt__(self, other):
        """Make it easy to compare two course states."""
        return self._d["priority"] < other["priority"]


class Course(BasePageExtension):
    """
    The course page extension represents and records a course in the catalog.

    This model should be used to record structured data about the course whereas the
    associated page object is where we record the less structured information to display on the
    page that presents the course.
    """

    effort = EffortField(
        time_units=defaults.TIME_UNITS,
        default_effort_unit=defaults.DEFAULT_EFFORT_UNIT,
        default_reference_unit=defaults.DEFAULT_REFERENCE_UNIT,
        max_length=80,
        blank=True,
        null=True,
    )
    duration = CompositeDurationField(
        time_units=defaults.TIME_UNITS,
        default_unit=defaults.DEFAULT_TIME_UNIT,
        max_length=80,
        blank=True,
        null=True,
    )

    PAGE = defaults.COURSES_PAGE

    class Meta:
        db_table = "richie_course"
        verbose_name = _("course")

    def __str__(self):
        """Human representation of a course."""
        return "{model}: {title}".format(
            model=self._meta.verbose_name.title(),
            title=self.extended_object.get_title(),
        )

    def create_page_role(self):
        """
        Create a new page role for the course with:
          - a user group to handle permissions for admins of this course,
          - a folder in Django Filer to store images related to this course,
          - all necessary permissions.
        """
        if not getattr(settings, "CMS_PERMISSION", False):
            return None

        # The page role is only created for draft courses
        if not self.extended_object.publisher_is_draft:
            return None

        # Don't do anything if it already exists
        page_role = PageRole.objects.filter(
            page=self.extended_object, role=defaults.ADMIN
        ).first()

        if page_role:
            return page_role

        # Create a role for admins of this course (which will automatically create a new
        # user group and a new Filer folder)
        page_role = PageRole.objects.create(
            page=self.extended_object, role=defaults.ADMIN
        )

        # Associate permissions as defined in settings:
        # - Create Django permissions
        page_role.group.permissions.set(
            get_permissions(defaults.COURSE_ADMIN_ROLE.get("django_permissions", []))
        )

        # - Create DjangoCMS page permissions
        PagePermission.objects.create(
            group_id=page_role.group_id,
            page=self.extended_object,
            **defaults.COURSE_ADMIN_ROLE.get("course_page_permissions", {}),
        )

        # - Create the Django Filer folder permissions
        FolderPermission.objects.create(
            folder_id=page_role.folder_id,
            group_id=page_role.group_id,
            **defaults.COURSE_ADMIN_ROLE.get("course_folder_permissions", {}),
        )
        return page_role

    def create_permissions_for_organization(self, organization):
        """
        Create page and folder permissions on the course page for the admin group of the
        organization passed in argument.
        """
        course_page_role = self.create_page_role()
        organization_page_role = organization.create_page_role()

        if organization_page_role is None or course_page_role is None:
            return

        # - Create DjangoCMS page permissions
        PagePermission.objects.create(
            group_id=organization_page_role.group_id,
            page_id=self.extended_object_id,
            **defaults.ORGANIZATION_ADMIN_ROLE.get("courses_page_permissions", {}),
        )

        # - Create the Django Filer folder permissions
        FolderPermission.objects.create(
            folder_id=course_page_role.folder_id,
            group_id=organization_page_role.group_id,
            **defaults.ORGANIZATION_ADMIN_ROLE.get("courses_folder_permissions", {}),
        )

    def get_persons(self, language=None):
        """
        Return the persons linked to the course via a person plugin in any of the
        placeholders on the course detail page, ranked by their `path` to respect
        the order in the persons tree.
        """
        language = language or translation.get_language()

        selector = "extended_object__person_plugins__cmsplugin_ptr"
        # pylint: disable=no-member
        filter_dict = {
            "{:s}__language".format(selector): language,
            "{:s}__placeholder__page".format(selector): self.extended_object,
        }
        # For a public course, we must filter out persons that are not published in
        # any language
        if self.extended_object.publisher_is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            Person.objects.filter(**filter_dict)
            .select_related("extended_object")
            .order_by("extended_object__node__path")
            .distinct()
        )

    def get_organizations(self, language=None):
        """
        Return the organizations linked to the course via an organization plugin in any
        of the placeholders on the course detail page, ranked by their `path` to respect
        the order in the organizations tree.
        """
        language = language or translation.get_language()

        selector = "extended_object__organization_plugins__cmsplugin_ptr"
        # pylint: disable=no-member
        filter_dict = {
            "{:s}__language".format(selector): language,
            "{:s}__placeholder__page".format(selector): self.extended_object,
        }
        # For a public course, we must filter out organizations that are not published in
        # any language
        if self.extended_object.publisher_is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            Organization.objects.filter(**filter_dict)
            .select_related("extended_object")
            .order_by("extended_object__node__path")
            .distinct()
        )

    def get_main_organization(self):
        """
        Return the main organization linked to the course via an organization plugin on the
        course detail page.

        Plugins are sortable by drag-and-drop in the plugin bar so we use this "position"
        information to return the first one as the main organization. We assume that only one
        placeholder holds organization plugins and return the first organization in order of
        position in this placeholder.
        """
        return (
            self.get_organizations()
            # .order_by("extended_object__organization_plugins__cmsplugin_ptr__position")
            .first()
        )

    def get_categories(self, language=None):
        """
        Return the categories linked to the course via a category plugin in any of the
        placeholders on the course detail page, ranked by their `path` to respect the
        order in the categories tree.
        """
        language = language or translation.get_language()

        selector = "extended_object__category_plugins__cmsplugin_ptr"
        # pylint: disable=no-member
        filter_dict = {
            "{:s}__language".format(selector): language,
            "{:s}__placeholder__page".format(selector): self.extended_object,
        }
        # For a public course, we must filter out categories that are not published in
        # any language
        if self.extended_object.publisher_is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            Category.objects.filter(**filter_dict)
            .select_related("extended_object")
            .order_by("extended_object__node__path")
            .distinct()
        )

    def get_root_to_leaf_category_pages(self):
        """
        Build a query retrieving all pages linked to this course's categories or one of their
        ancestors excluding the meta category itself.

        This is useful to build the course Elasticsearch index. When a category is linked to a
        course, we associate all its ancestors (excluding the meta category) to the course so
        that a search for courses in an ancestor category also returns the course.
        """
        category_query = self.get_categories()

        # 1. We want the pages directly related to a category of the course
        page_query = Page.objects.filter(publisher_draft__category__in=category_query)
        # 2. We want the pages related to one of the ancestors of the categories of the course
        for category in category_query.select_related(
            "public_extension__extended_object"
        ):
            ancestor_query = (
                category.public_extension.extended_object.get_ancestor_pages()
            )
            # Don't include the meta category as it materializes a "filter bank" and not a
            # search option
            page_query = page_query | ancestor_query.filter(
                node__parent__cms_pages__category__isnull=False
            )
        return page_query.distinct()

    def get_course_runs(self):
        """
        Returns a query yielding the course runs related to the course. They may be direct
        children of the course or children of a snapshot of the course.

        For a draft course instance, the related draft course runs are retrieved.
        For a public course instance, the related public course runs are retrieved, but only
        those that are currently published in at least one language*.

        (*) The catch here is that a course run could have been published and then unpublished
            for all languages. The public instance is created the first time a draft page is
            published. Of course, this public instance will still exist if the object is then
            unpublished for all languages...
            Said differently, a page can have a public version of itself in database but not be
            currently published in any language.
        """
        node = self.extended_object.node
        is_draft = self.extended_object.publisher_is_draft
        filter_dict = {
            "extended_object__publisher_is_draft": is_draft,
            "extended_object__node__path__startswith": node.path,
            "extended_object__node__depth__gt": node.depth,
        }
        # For a public course, we must filter out course runs that are not published in
        # any language
        if is_draft is False:
            filter_dict["extended_object__title_set__published"] = True

        return (
            CourseRun.objects.filter(**filter_dict)
            .order_by("extended_object__node__path")
            .distinct()
        )

    def get_course_runs_for_language(self, language=None):
        """
        Returns a query yielding the course runs related to the course for the current
        language (or the language passed in arguments). They may be direct children of
        the course or children of a snapshot of the course.
        """
        language = language or translation.get_language()
        course_runs = self.get_course_runs()

        # For a public course, we must filter out course runs that are not published in
        # the language
        if self.extended_object.publisher_is_draft is False:
            course_runs = course_runs.filter(
                extended_object__title_set__language=language,
                extended_object__title_set__published=True,
            )
        else:
            course_runs = course_runs.filter(
                extended_object__title_set__language=language
            )

        return course_runs

    @property
    def state(self):
        """
        The state of the course carrying information on what to display on a course glimpse.

        The game is to find the highest priority state for this course among its course runs.
        """
        # The default state is for a course that has no course runs
        best_state = CourseState(6)

        for course_run in self.get_course_runs_for_language().only(
            "start", "end", "enrollment_start", "enrollment_end"
        ):
            state = course_run.state
            if state < best_state:
                best_state = state
            if state == 0:
                # We found the best state, don't waste more time
                break

        return best_state

    def save(self, *args, **kwargs):
        """
        Enforce validation each time an instance is saved
        """
        self.full_clean()
        super().save(*args, **kwargs)


class CourseRun(BasePageExtension):
    """
    The course run represents and records the occurence of a course between a start
    and an end date.
    """

    resource_link = models.URLField(_("Resource link"), blank=True, null=True)
    start = models.DateTimeField(_("course start"), blank=True, null=True)
    end = models.DateTimeField(_("course end"), blank=True, null=True)
    enrollment_start = models.DateTimeField(
        _("enrollment start"), blank=True, null=True
    )
    enrollment_end = models.DateTimeField(_("enrollment end"), blank=True, null=True)
    languages = MultiSelectField(
        max_choices=50,
        max_length=255,  # MySQL does not allow max_length > 255
        # Language choices are made lazy so that we can override them in our tests.
        # When set directly, they are evaluated too early and can't be changed with the
        # "override_settings" utility.
        choices=lazy(lambda: ALL_LANGUAGES, tuple)(),
        help_text=_("The list of languages in which the course content is available."),
    )

    PAGE = defaults.COURSERUNS_PAGE

    class Meta:
        db_table = "richie_course_run"
        verbose_name = _("course run")

    def __str__(self):
        """Human representation of a course run."""
        start = "{:%y/%m/%d %H:%M} - ".format(self.start) if self.start else ""
        return "{start:s}{course:s}".format(
            course=self.extended_object.get_title(), start=start
        )

    # pylint: disable=arguments-differ
    def save(self, *args, **kwargs):
        """
        Enforce validation each time an instance is saved.
        """
        self.full_clean()
        super().save(*args, **kwargs)

    # pylint: disable=too-many-return-statements
    @staticmethod
    def compute_state(start, end, enrollment_start, enrollment_end):
        """
        Compute at the current time the state of a course run that would have the dates
        passed in argument.

        A static method not using the instance allows to call it with an Elasticsearch result.
        """
        if not start or not enrollment_start:
            return CourseState(6)

        # course run end dates are not required and should default to forever
        # e.g. a course run with no end date is presumed to be always on-going
        end = end or MAX_DATE
        enrollment_end = enrollment_end or end

        now = timezone.now()
        if start < now:
            if end > now:
                if enrollment_end > now:
                    # ongoing open
                    return CourseState(0, enrollment_end)
                # ongoing closed
                return CourseState(4)
            # archived
            return CourseState(5)
        if enrollment_start > now:
            # future not yet open
            return CourseState(2, start)
        if enrollment_end > now:
            # future open
            return CourseState(1, start)
        # future already closed
        return CourseState(3)

    @property
    def state(self):
        """Return the state of the course run at the current time."""
        return self.compute_state(
            self.start, self.end, self.enrollment_start, self.enrollment_end
        )

    def get_course(self):
        """Get the course for this course run."""
        nodes = self.extended_object.node.get_ancestors()
        return Course.objects.get(
            # Joining on `cms_pages` generate duplicates for courses that are under a parent page
            # when this page exists both in draft and public versions. We need to exclude the
            # parent public page to avoid this duplication
            Q(
                extended_object__node__parent__cms_pages__publisher_is_draft=True
            )  # course has a parent
            | Q(extended_object__node__parent__isnull=True),  # course has no parent
            # Target courses that are ancestors of the course run
            extended_object__node__in=nodes,
            # Exclude snapshots
            extended_object__node__parent__cms_pages__course__isnull=True,  # exclude snapshots
            # Get the course in the same version as the course run
            extended_object__publisher_is_draft=self.extended_object.publisher_is_draft,
        )


class CoursePluginModel(PagePluginMixin, CMSPlugin):
    """
    Course plugin model handles the relation from CoursePlugin
    to their Course instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="course_plugins",
        limit_choices_to={
            # Joining on `cms_pages` generate duplicates for courses that are under a parent page
            # when this page exists both in draft and public versions. We need to exclude the
            # parent public page to avoid this duplication with the first condition.
            # The second condition makes sure the parent is not a course to exclude snapshots.
            # The third condition filters out public course.
            # The fourth condition makes sure only courses show up.
            "node__parent__cms_pages__publisher_is_draft": True,  # exclude course published parent
            "node__parent__cms_pages__course__isnull": True,  # limit to course - no snapshot
            "publisher_is_draft": True,  # plugins work with draft instances
            "course__isnull": False,  # limit to pages linked to a course object
        },
    )

    class Meta:
        db_table = "richie_course_plugin"
        verbose_name = _("course plugin")

    def __str__(self):
        """Human representation of a page plugin"""
        return "{model:s}: {id:d}".format(
            model=self._meta.verbose_name.title(), id=self.id
        )


class Licence(models.Model):
    """
    Licence model.

    Instances of this models should only be created by administrators.
    """

    name = models.CharField(_("name"), max_length=200)
    logo = FilerImageField(
        verbose_name=_("logo"), on_delete=models.PROTECT, related_name="licence"
    )
    url = models.CharField(_("url"), blank=True, max_length=255)
    content = models.TextField(_("content"), blank=False, default="")

    class Meta:
        db_table = "richie_licence"
        verbose_name = _("licence")

    def __str__(self):
        """Human representation of a person title"""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.name
        )


class LicencePluginModel(CMSPlugin):
    """
    Licence plugin model.
    """

    licence = models.ForeignKey(Licence, on_delete=models.CASCADE)
    description = models.TextField(_("description"), blank=True, default="")

    class Meta:
        db_table = "richie_licence_plugin"
        verbose_name = _("licence plugin")

    def __str__(self):
        """Human representation of a person plugin"""
        return "{model:s}: {name:s}".format(
            model=self._meta.verbose_name.title(), name=self.licence.name
        )


extension_pool.register(Course)
extension_pool.register(CourseRun)



"""
Courses application admin
"""
from django.conf.urls import url
from django.contrib import admin
from django.core.exceptions import PermissionDenied
from django.db import models as django_models
from django.db import transaction
from django.http import HttpResponseBadRequest, HttpResponseForbidden, JsonResponse
from django.utils.decorators import method_decorator
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.http import require_POST

from cms.admin.placeholderadmin import FrontendEditableAdminMixin
from cms.api import Page
from cms.extensions import PageExtensionAdmin
from cms.utils.admin import jsonify_request

from ..core.admin import link_field
from . import models
from .fields import CourseRunSplitDateTimeField
from .forms import AdminCategoryForm, AdminLicenceForm
from .helpers import snapshot_course
from .widgets import CourseRunSplitDateTimeWidget

REQUIRE_POST = method_decorator(require_POST)


class CategoryAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the Category model"""

    form = AdminCategoryForm


class CourseAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the Course model"""

    list_display = ["title"]
    frontend_editable_fields = ("effort", "duration")

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the course title as a read-only field from the related page
        """
        return obj.extended_object.get_title()

    def get_urls(self):
        """
        Add admin URL to trigger the snapshot of a course.
        """
        url_patterns = super().get_urls()
        return [
            url(
                r"^(?P<course_id>[0-9]+)/snapshot/$",
                self.admin_site.admin_view(self.snapshot),
                name="cms_course_snapshot",
            )
        ] + url_patterns

    @REQUIRE_POST
    @transaction.atomic
    def snapshot(self, request, course_id, *args, **kwargs):
        """
        Snapshotting a course is making a copy of the course page with all its permissions and
        extensions, and placing the copy as the first child of the page being snapshotted, then
        moving all the course run of the page being snapshotted as children of the new snapshot
        so we keep record of the course as it was when these course runs were played.
        """
        try:
            page = Page.objects.select_related("node__site").get(
                publisher_is_draft=True, course__id=course_id
            )
        except Page.DoesNotExist:
            return jsonify_request(
                HttpResponseBadRequest(force_text(_("Course could not be found.")))
            )

        try:
            new_page = snapshot_course(page, request.user, simulate_only=False)
        except PermissionDenied as context:
            return jsonify_request(HttpResponseForbidden(force_text(context)))

        return JsonResponse({"id": new_page.course.id})


class CourseRunAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the CourseRun model"""

    list_display = ["title"]
    frontend_editable_fields = (
        "languages",
        "resource_link",
        "start",
        "end",
        "enrollment_start",
        "enrollment_end",
    )
    formfield_overrides = {
        django_models.DateTimeField: {
            "form_class": CourseRunSplitDateTimeField,
            "widget": CourseRunSplitDateTimeWidget,
        }
    }

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the course title as a read-only field from the related page
        """
        return obj.extended_object.get_title()


class OrganizationAdmin(PageExtensionAdmin):
    """
    Admin class for the Organization model
    """

    list_display = ["title", "code"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Get the page title from the related page
        """
        return obj.extended_object.get_title()


class PageRoleAdmin(admin.ModelAdmin):
    """
    Admin class for the PageRole model.
    """

    fields = [
        "role",
        link_field("page"),
        link_field("group", anchor=_("See user group")),
        link_field("folder", anchor=_("See filer folder")),
    ]
    list_display = ["id", "role", "page"]
    list_filter = ["role"]
    search_fields = ("id", "page__title_set__title")

    def get_fieldsets(self, request, obj=None):
        fieldsets = super(PageRoleAdmin, self).get_fieldsets(request, obj)
        if not obj:
            fieldsets[0][1]["fields"] = ["role", "page"]
        return fieldsets

    def get_readonly_fields(self, request, obj=None):
        """Don't allow modifying the object once created."""
        if obj:
            return self.fields
        return []


class PersonAdmin(PageExtensionAdmin):
    """Admin class for the Person model"""

    list_display = ["title"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the person page title as a read-only field from the related page
        """
        return obj.extended_object.get_title()


class LicenceAdmin(admin.ModelAdmin):
    """
    Admin class for the Licence model
    """

    list_display = ["name"]
    form = AdminLicenceForm


admin.site.register(models.Category, CategoryAdmin)
admin.site.register(models.Course, CourseAdmin)
admin.site.register(models.CourseRun, CourseRunAdmin)
admin.site.register(models.Licence, LicenceAdmin)
admin.site.register(models.Organization, OrganizationAdmin)
admin.site.register(models.PageRole, PageRoleAdmin)
admin.site.register(models.Person, PersonAdmin)



"""
Default settings for the Richie courses app.

If you use Django Configuration for your settings, you can use our mixin to import these
default settings:
    ```
    from configurations import Configuration
    from richie.apps.courses.settings.mixins import RichieCoursesConfigurationMixin

    class MyConfiguration(RichieCoursesConfigurationMixin, Configuration):
        ...
    ```

Otherwise, you can just use the usual Django pattern in your settings.py file:
    ```
    from richie.apps.courses.settings import *
    ```
"""
from django.utils.translation import ugettext_lazy as _

# Easy Thumbnails
THUMBNAIL_PROCESSORS = (
    "easy_thumbnails.processors.colorspace",
    "easy_thumbnails.processors.autocrop",
    "filer.thumbnail_processors.scale_and_crop_with_subject_location",
    "easy_thumbnails.processors.filters",
    "easy_thumbnails.processors.background",
)

# Django CMS
CMS_TEMPLATES = (
    ("courses/cms/course_detail.html", _("Course page")),
    ("courses/cms/course_run_detail.html", _("Course run page")),
    ("courses/cms/organization_list.html", _("Organization list")),
    ("courses/cms/organization_detail.html", _("Organization page")),
    ("courses/cms/category_list.html", _("Category list")),
    ("courses/cms/category_detail.html", _("Category page")),
    ("courses/cms/blogpost_list.html", _("Blog post list")),
    ("courses/cms/blogpost_detail.html", _("Blog post page")),
    ("courses/cms/person_detail.html", _("Person page")),
    ("courses/cms/person_list.html", _("Person list")),
    ("search/search.html", _("Search")),
    ("richie/child_pages_list.html", _("List of child pages")),
    ("richie/homepage.html", _("Homepage")),
    ("richie/single_column.html", _("Single column")),
)

CMS_PLACEHOLDER_CONF = {
    # Homepage
    "richie/homepage.html maincontent": {
        "name": _("Main content"),
        "plugins": ["LargeBannerPlugin", "SectionPlugin"],
        "child_classes": {
            "SectionPlugin": [
                "BlogPostPlugin",
                "CoursePlugin",
                "CategoryPlugin",
                "LinkPlugin",
                "OrganizationPlugin",
                "PersonPlugin",
            ]
        },
    },
    # Single column page
    "richie/single-column.html maincontent": {
        "name": _("Main content"),
        "excluded_plugins": ["CKEditorPlugin", "GoogleMapPlugin"],
        "parent_classes": {
            "BlogPostPlugin": ["SectionPlugin"],
            "CategoryPlugin": ["SectionPlugin"],
            "CoursePlugin": ["SectionPlugin"],
            "OrganizationPlugin": ["SectionPlugin"],
            "PersonPlugin": ["SectionPlugin"],
        },
        "child_classes": {
            "SectionPlugin": [
                "BlogPostPlugin",
                "CategoryPlugin",
                "CoursePlugin",
                "LinkPlugin",
                "OrganizationPlugin",
                "PersonPlugin",
            ]
        },
    },
    # Course detail
    "courses/cms/course_detail.html course_cover": {
        "name": _("Cover"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/course_detail.html course_introduction": {
        "name": _("Catch phrase"),
        "plugins": ["PlainTextPlugin"],
        "limits": {"PlainTextPlugin": 1},
    },
    "courses/cms/course_detail.html course_teaser": {
        "name": _("Teaser"),
        "plugins": ["VideoPlayerPlugin", "SimplePicturePlugin"],
        "limits": {"VideoPlayerPlugin": 1, "SimplePicturePlugin": 1},
    },
    "courses/cms/course_detail.html course_description": {
        "name": _("About the course"),
        "plugins": ["CKEditorPlugin"],
    },
    "courses/cms/course_detail.html course_skills": {
        "name": _("What you will learn"),
        "plugins": ["CKEditorPlugin"],
    },
    "courses/cms/course_detail.html course_format": {
        "name": _("Format"),
        "plugins": ["PlainTextPlugin"],
    },
    "courses/cms/course_detail.html course_prerequisites": {
        "name": _("Prerequisites"),
        "plugins": ["PlainTextPlugin"],
    },
    "courses/cms/course_detail.html course_team": {
        "name": _("Team"),
        "plugins": ["PersonPlugin"],
    },
    "courses/cms/course_detail.html course_plan": {
        "name": _("Plan"),
        "plugins": ["CKEditorPlugin"],
    },
    "courses/cms/course_detail.html course_information": {
        "name": _("Complementary information"),
        "plugins": ["SectionPlugin"],
        "parent_classes": {
            "CKEditorPlugin": ["SectionPlugin"],
            "SimplePicturePlugin": ["SectionPlugin"],
        },
        "child_classes": {"SectionPlugin": ["CKEditorPlugin", "SimplePicturePlugin"]},
    },
    "courses/cms/course_detail.html course_license_content": {
        "name": _("License for the course content"),
        "plugins": ["LicencePlugin"],
        "limits": {"LicencePlugin": 1},
    },
    "courses/cms/course_detail.html course_license_participation": {
        "name": _("License for the content created by course participants"),
        "plugins": ["LicencePlugin"],
        "limits": {"LicencePlugin": 1},
    },
    "courses/cms/course_detail.html course_categories": {
        "name": _("Categories"),
        "plugins": ["CategoryPlugin"],
    },
    "courses/cms/course_detail.html course_icons": {
        "name": _("Icon"),
        "plugins": ["CategoryPlugin"],
        "limits": {"CategoryPlugin": 1},
    },
    "courses/cms/course_detail.html course_organizations": {
        "name": _("Organizations"),
        "plugins": ["OrganizationPlugin"],
    },
    "courses/cms/course_detail.html course_assessment": {
        "name": _("Assessment and Certification"),
        "plugins": ["PlainTextPlugin"],
    },
    # Organization detail
    "courses/cms/organization_detail.html banner": {
        "name": _("Banner"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/organization_detail.html logo": {
        "name": _("Logo"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/organization_detail.html description": {
        "name": _("Description"),
        "plugins": ["CKEditorPlugin"],
        "limits": {"CKEditorPlugin": 1},
    },
    # Category detail
    "courses/cms/category_detail.html banner": {
        "name": _("Banner"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/category_detail.html logo": {
        "name": _("Logo"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/category_detail.html icon": {
        "name": _("Icon"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/category_detail.html description": {
        "name": _("Description"),
        "plugins": ["CKEditorPlugin"],
        "limits": {"CKEditorPlugin": 1},
    },
    # Person detail
    "courses/cms/person_detail.html categories": {
        "name": _("Categories"),
        "plugins": ["CategoryPlugin"],
    },
    "courses/cms/person_detail.html portrait": {
        "name": _("Portrait"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/person_detail.html bio": {
        "name": _("Bio"),
        "plugins": ["PlainTextPlugin"],
        "limits": {"PlainTextPlugin": 1},
    },
    "courses/cms/person_detail.html organizations": {
        "name": _("Organizations"),
        "plugins": ["OrganizationPlugin"],
    },
    # Blog page detail
    "courses/cms/blogpost_detail.html author": {
        "name": _("Author"),
        "plugins": ["PersonPlugin"],
        "limits": {"PersonPlugin": 1},
    },
    "courses/cms/blogpost_detail.html categories": {
        "name": _("Categories"),
        "plugins": ["CategoryPlugin"],
    },
    "courses/cms/blogpost_detail.html cover": {
        "name": _("Cover"),
        "plugins": ["SimplePicturePlugin"],
        "limits": {"SimplePicturePlugin": 1},
    },
    "courses/cms/blogpost_detail.html excerpt": {
        "name": _("Excerpt"),
        "plugins": ["PlainTextPlugin"],
        "limits": {"PlainTextPlugin": 1},
    },
    "courses/cms/blogpost_detail.html body": {
        "name": _("Body"),
        "excluded_plugins": ["CKEditorPlugin", "GoogleMapPlugin"],
    },
}

# Main CKEditor configuration
CKEDITOR_SETTINGS = {
    "language": "{{ language }}",
    "skin": "moono-lisa",
    "toolbarCanCollapse": False,
    "contentsCss": "/static/richie/css/ckeditor.css",
    # Enabled showblocks as default behavior
    "startupOutlineBlocks": True,
    # Enable some plugins
    # 'extraPlugins': 'codemirror',
    # Disable element filter to enable full HTML5, also this will let
    # append any code, even bad syntax and malicious code, so be careful
    "removePlugins": "stylesheetparser",
    "allowedContent": True,
    # Image plugin options
    "image_prefillDimensions": False,
    # Justify text using shortand class names
    "justifyClasses": ["text-left", "text-center", "text-right"],
    # Default toolbar configurations for djangocms_text_ckeditor
    "toolbar": "CMS",
    "toolbar_CMS": [
        ["Undo", "Redo"],
        ["cmsplugins", "-", "ShowBlocks"],
        ["Format", "Styles"],
        ["RemoveFormat"],
        ["Maximize"],
        "/",
        ["Bold", "Italic", "Underline", "-", "Subscript", "Superscript"],
        ["JustifyLeft", "JustifyCenter", "JustifyRight"],
        ["Link", "Unlink"],
        ["NumberedList", "BulletedList", "-", "HorizontalRule"],
        ["Source"],
    ],
}
# Share the same configuration for djangocms_text_ckeditor field and derived
# CKEditor widgets/fields
CKEDITOR_SETTINGS["toolbar_HTMLField"] = CKEDITOR_SETTINGS["toolbar_CMS"]

# CKEditor configuration for basic formatting
CKEDITOR_BASIC_CONFIGURATION = {
    "language": "{{ language }}",
    "skin": "moono-lisa",
    "toolbarCanCollapse": False,
    "contentsCss": "/static/css/ckeditor.css",
    # Only enable following tag definitions
    "allowedContent": ["p", "b", "i", "a[href]"],
    # Enabled showblocks as default behavior
    "startupOutlineBlocks": True,
    # Default toolbar configurations for djangocms_text_ckeditor
    "toolbar": "HTMLField",
    "toolbar_HTMLField": [["Undo", "Redo"], ["Bold", "Italic"], ["Link", "Unlink"]],
}

# CKEditor configuration for formatting limited to:
# paragraph, bold, italic and numbered or bulleted lists.
CKEDITOR_LIMITED_CONFIGURATION = {
    "language": "{{ language }}",
    "skin": "moono-lisa",
    "toolbarCanCollapse": False,
    "contentsCss": "/static/css/ckeditor.css",
    # Only enable following tag definitions
    "allowedContent": ["p", "b", "i", "ol", "ul", "li"],
    # Enabled showblocks as default behavior
    "startupOutlineBlocks": True,
    # Default toolbar configurations for djangocms_text_ckeditor
    "toolbar": "HTMLField",
    "toolbar_HTMLField": [
        ["Undo", "Redo"],
        ["Bold", "Italic"],
        ["Link", "Unlink"],
        ["NumberedList", "BulletedList", "-"],
    ],
}

# Additional LinkPlugin templates. Note how choice value is just a keyword
# instead of full template path. Value is used inside a path formatting
# such as "templates/djangocms_link/VALUE/link.html"
DJANGOCMS_LINK_TEMPLATES = [("button-caesura", _("Button caesura"))]

DJANGOCMS_VIDEO_TEMPLATES = [("full-width", _("Full width"))]

# Richie plugins

RICHIE_PLAINTEXT_MAXLENGTH = {"course_introduction": 200, "bio": 150, "excerpt": 200}

RICHIE_SIMPLETEXT_CONFIGURATION = [
    {
        "placeholders": ["course_skills", "course_plan"],
        "ckeditor": "CKEDITOR_LIMITED_CONFIGURATION",
    },
    {
        "placeholders": ["course_description"],
        "ckeditor": "CKEDITOR_LIMITED_CONFIGURATION",
        "max_length": 1200,
    },
]

RICHIE_SIMPLEPICTURE_PRESETS = {
    # Formatting images for the courses search index
    "cover": {
        "src": {"size": (300, 170), "crop": "smart"},
        "srcset": [
            {
                "options": {"size": (300, 170), "crop": "smart", "upscale": True},
                "descriptor": "300w",
            },
            {
                "options": {"size": (600, 340), "crop": "smart", "upscale": True},
                "descriptor": "600w",
            },
            {
                "options": {"size": (900, 560), "crop": "smart", "upscale": True},
                "descriptor": "900w",
            },
        ],
        "sizes": "300px",
    },
    "icon": {
        "src": {"size": (60, 60), "crop": "smart"},
        "srcset": [
            {
                "options": {"size": (60, 60), "crop": "smart", "upscale": True},
                "descriptor": "60w",
            },
            {
                "options": {"size": (120, 120), "crop": "smart", "upscale": True},
                "descriptor": "120w",
            },
            {
                "options": {"size": (180, 180), "crop": "smart", "upscale": True},
                "descriptor": "180w",
            },
        ],
        "sizes": "60px",
    },
}



"""HTML sitemap plugin models."""
from django.db import models
from django.utils.text import Truncator
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from cms.models.pluginmodel import CMSPlugin


class HTMLSitemapPage(CMSPlugin):
    """
    HTML Sitemap page plugin model to define a specific subtree to be included in the sitemap
    and configure which pages should be excluded or not:
        - pages requiring login,
        - maximum nesting depth,
        - include or not pages that are excluded from navigation,
        - include the root page and its descendants or just its descendants.
    """

    root_page = models.ForeignKey(
        to=Page,
        limit_choices_to={"publisher_is_draft": True},
        related_name="sitemaps",
        on_delete=models.PROTECT,
        verbose_name=_("root page"),
        help_text=_(
            "This page will be at the root of your sitemap (or its children if the "
            '"include root page" flag is unticked).'
        ),
        blank=True,
        null=True,
    )
    max_depth = models.PositiveIntegerField(
        verbose_name=_("max depth"),
        help_text=_(
            "Limit the level of nesting that your sitemap will contain below this page. "
            "An empty field or 0 equals to no limit."
        ),
        blank=True,
        null=True,
    )
    in_navigation = models.BooleanField(
        default=False,
        verbose_name=_("in navigation"),
        help_text=_(
            "Tick to exclude from sitemap the pages that are excluded from navigation."
        ),
    )
    include_root_page = models.BooleanField(
        default=True,
        verbose_name=_("include root page"),
        help_text=_(
            "Tick to include the root page and its descendants. "
            "Untick to include only its descendants."
        ),
    )

    class Meta:
        verbose_name = _("HTML Sitemap")
        verbose_name_plural = _("HTML Sitemaps")

    def __str__(self):
        return (
            Truncator(self.root_page.get_title()).words(6, truncate="...")
            if self.root_page
            else str(_("Sitemap"))
        )



"""
Large banner plugin field choices
"""
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

LARGEBANNER_TEMPLATES = getattr(
    settings,
    "RICHIE_LARGEBANNER_TEMPLATES",
    [
        ("richie/large_banner/large_banner.html", _("Default")),
        ("richie/large_banner/hero-intro.html", _("Hero introduction")),
    ],
)



"""
Declare and configure the model for the person application
"""
from django.apps import apps
from django.db import models
from django.db.models import Prefetch
from django.utils import translation
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models import Title
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension, PagePluginMixin
from ..defaults import PERSONS_PAGE


class Person(BasePageExtension):
    """
    The person page extension represents and records people information.
    It could be a course or news article author.

    This model should be used to record structured data about the person whereas the
    associated page object is where we record the less structured information to display on the
    page to present the person.
    """

    PAGE = PERSONS_PAGE

    class Meta:
        db_table = "richie_person"
        verbose_name = _("person")

    def __str__(self):
        """Human representation of a person."""
        return "{model}: {name}".format(
            name=self.extended_object.get_title(), model=self._meta.verbose_name.title()
        )

    def save(self, *args, **kwargs):
        """
        Enforce validation on each instance save
        """
        self.full_clean()
        super().save(*args, **kwargs)

    def get_courses(self, language=None):
        """
        Return a query to get the courses related to this person ie for which a plugin for
        this person is linked to the course page via any placeholder.
        """
        is_draft = self.extended_object.publisher_is_draft
        person = self if is_draft else self.draft_extension
        language = language or translation.get_language()

        bfs = (
            "extended_object__placeholders__cmsplugin__courses_personpluginmodel__page"
        )
        filter_dict = {
            "extended_object__node__parent__cms_pages__course__isnull": True,  # exclude snapshots
            "extended_object__publisher_is_draft": is_draft,
            "extended_object__placeholders__cmsplugin__language": language,
            bfs: person.extended_object,
        }

        course_model = apps.get_model(app_label="courses", model_name="course")
        # pylint: disable=no-member
        return (
            course_model.objects.filter(**filter_dict)
            .select_related("extended_object")
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="prefetched_titles",
                    queryset=Title.objects.filter(language=language),
                )
            )
            .distinct()
        )

    def get_blogposts(self, language=None):
        """
        Return a query to get the blogposts this person wrote ie for which a
        plugin for this person is linked to the blogpost page via any placeholder.
        """
        is_draft = self.extended_object.publisher_is_draft
        person = self if is_draft else self.draft_extension
        language = language or translation.get_language()

        bfs = (
            "extended_object__placeholders__cmsplugin__courses_personpluginmodel__page"
        )
        filter_dict = {
            "extended_object__publisher_is_draft": is_draft,
            "extended_object__placeholders__cmsplugin__language": language,
            bfs: person.extended_object,
        }

        blogpost_model = apps.get_model(app_label="courses", model_name="blogpost")
        # pylint: disable=no-member
        return (
            blogpost_model.objects.filter(**filter_dict)
            .select_related("extended_object")
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="prefetched_titles",
                    queryset=Title.objects.filter(language=language),
                )
            )
            .distinct()
        )


class PersonPluginModel(PagePluginMixin, CMSPlugin):
    """
    Person plugin model handles the relation from PersonPlugin
    to their Person instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        limit_choices_to={"publisher_is_draft": True, "person__isnull": False},
        related_name="person_plugins",
    )

    class Meta:
        db_table = "richie_person_plugin"
        verbose_name = _("person plugin")

    def __str__(self):
        """Human representation of a person plugin"""
        return "{model:s}: {id:d}".format(
            model=self._meta.verbose_name.title(), id=self.id
        )


extension_pool.register(Person)



"""Helpers for the demo app of the Richie project."""
from richie.apps.courses.factories import CategoryFactory


# pylint: disable=too-many-arguments
def create_categories(
    children=None,
    color=None,
    fill_banner=True,
    fill_description=True,
    fill_icon=False,
    fill_logo=True,
    page_in_navigation=True,
    page_title=None,
    page_parent=None,
    page_reverse_id=None,
    should_publish=True,
):
    """
    Create the category tree from the SUBJECTS dictionary.


    Arguments:
        info (List): definition of the category tree to create in the following format:

            {
                "title": "Subject",
                "children": [
                    {
                        "title": "Computer science",
                        "children": [
                            {"title": "Coding"},
                            {"title": "Security"},
                        ],
                    },
                    {"title": "Languages"},
                ],
            }

        page (cms.models.pagemodel.Page): Instance of a Page below which the category
            tree is created.

    Returns:
        generator[courses.models.Category]: yield only the leaf categories of the created tree.

    """
    category = CategoryFactory(
        color=color,
        fill_banner=fill_banner,
        fill_description=fill_description,
        fill_icon=fill_icon,
        fill_logo=fill_logo,
        page_title=page_title,
        page_reverse_id=page_reverse_id,
        page_in_navigation=page_in_navigation,
        page_parent=page_parent,
        should_publish=should_publish,
    )

    if children:
        for child_info in children:
            kwargs = {
                "fill_banner": fill_banner,
                "fill_description": fill_description,
                "fill_icon": fill_icon,
                "fill_logo": fill_logo,
                "page_parent": category.extended_object,
                "should_publish": should_publish,
            }
            kwargs.update(child_info)
            yield from create_categories(**kwargs)
    else:
        # we only return leaf categories (no children)
        yield category



"""Create_demo_site management command."""
import logging

from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand

from richie.apps.core.helpers import recursive_page_creation

from ...defaults import PAGES_INFO

logger = logging.getLogger("richie.commands.core.richie_init")


class Command(BaseCommand):
    """Create the minimum site structure required by Richie."""

    help = __doc__

    def handle(self, *args, **options):
        """Call the `richie_init` function."""
        site = Site.objects.get(id=1)
        recursive_page_creation(site, PAGES_INFO)



"""
Common utilities related to our indexers. For use in our indexers and related settings,
or as helpers for users of the project.
"""
from django.utils.module_loading import import_string


class IndicesList:
    """
    This utility class allow configuring indexers from the settings and still benefit from
    an easy access and handling of the instantiated indexers. For example:

        >>> indices = IndicesList(
        ...     "courses": "richie.apps.search.indexers.courses.CoursesIndexer",
        ...     "categories": "richie.apps.search.indexers.categories.CategoriesIndexer",
        ... )
        >>> courses_indexer = indices.courses
        >>> for indexer in indexers:
        ...     form = indexer.form()....
    """

    def __init__(self, **kwargs):
        """Record any dotted path to indices to allow access by property or iteration."""
        self.index_map = kwargs

    def __getattr__(self, item):
        """Allow accessing indices via properties (e.g. indices.courses)."""
        return import_string(self.index_map[item])

    def __iter__(self):
        """Allow iterating through indices (e.g. for index in indices)."""
        return iter(
            [import_string(dotted_path) for dotted_path in self.index_map.values()]
        )


def slice_string_for_completion(string):
    """
    Split a string in significant parts for use in completion.
    Example: "University of Paris 13" => "University of Paris 13", "of Paris 13", "Paris 13", "13"

    This is useful to enable autocompletion starting from any part of a name. If we just use the
    name directly in the ES completion type, it will only return options that match on the first
    characters of the whole string, which is not always suitable.
    """
    parts = [part for part in string.split(" ") if part != ""]
    return [" ".join(parts[index:]) for index, _ in enumerate(parts)]



# Generated by Django 2.1.10 on 2019-07-17 06:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("courses", "0004_auto_20190619_1630")]

    operations = [
        migrations.AddField(
            model_name="category",
            name="color",
            field=models.CharField(blank=True, max_length=10, null=True),
        )
    ]



"""
Declare and configure the models for the blog part
"""
from django.db import models
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension, PagePluginMixin
from ..defaults import BLOGPOSTS_PAGE


class BlogPost(BasePageExtension):
    """
    The blogpost extension represents and records a blog article.
    """

    PAGE = BLOGPOSTS_PAGE

    class Meta:
        db_table = "richie_blog_post"
        ordering = ["-pk"]
        verbose_name = _("blog post")

    def __str__(self):
        """Human representation of a blogpost"""
        return "{model}: {name}".format(
            name=self.extended_object.get_title(), model=self._meta.verbose_name.title()
        )


class BlogPostPluginModel(PagePluginMixin, CMSPlugin):
    """
    BlogPost plugin model handles the relation between the BlogPostPluginModel
    and its blogpost instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="blogpost_plugins",
        limit_choices_to={"publisher_is_draft": True, "blogpost__isnull": False},
    )

    class Meta:
        db_table = "richie_blog_post_plugin"
        verbose_name = _("blog post plugin")

    def __str__(self):
        """Human representation of a blogpost plugin"""
        return "{model:s}: {id:d}".format(
            model=self._meta.verbose_name.title(), id=self.id
        )


extension_pool.register(BlogPost)



"""Signals to update the Elasticsearch indexes when page modifications are published."""
from django.apps import AppConfig


class SearchConfig(AppConfig):
    """Configuration class for the search app."""

    name = "richie.apps.search"
    verbose_name = "Richie search app"

    def ready(self):
        """Register signals to update the Elasticsearch indexes."""
        from cms.signals import post_publish
        from .signals import on_page_publish

        post_publish.connect(on_page_publish, dispatch_uid="search_post_publish")



"""
Simple text plugin models
"""
from django.db import models
from django.utils.encoding import force_text, python_2_unicode_compatible
from django.utils.html import strip_tags
from django.utils.text import Truncator
from django.utils.translation import ugettext_lazy as _

from cms.models.pluginmodel import CMSPlugin
from djangocms_text_ckeditor.html import clean_html


@python_2_unicode_compatible
class SimpleText(CMSPlugin):
    """
    Simple text plugin model.

    Aim to be used instead of TextPlugin which produces orphan plugin item that
    broke some behavior with constraints from ``CMS_PLACEHOLDER_CONF``.
    """

    body = models.TextField(_("body"))

    def __str__(self):
        return Truncator(strip_tags(self.body)).words(6, truncate="...")

    def __init__(self, *args, **kwargs):
        super(SimpleText, self).__init__(*args, **kwargs)
        self.body = force_text(self.body)

    # pylint: disable=arguments-differ
    def save(self, *args, **kwargs):
        # Clean HTML from potential XSS content
        self.body = clean_html(self.body, full=False)

        super(SimpleText, self).save(*args, **kwargs)



"""Define a custom manager for page extensions"""
from django.db import models
from django.utils import translation

from cms.extensions import PageExtension


class PageExtensionQuerySet(models.QuerySet):
    """
    Add custom filters to the default Django queryset for page extensions.
    """

    def drafts(self):
        """
        Custom filter to get only the draft page extension instances based on the status of the
        linked DjangoCMS page.
        """
        return self.filter(extended_object__publisher_is_draft=True)

    def published(self):
        """
        Custom filter to get only the published page extension instances based on the status of the
        linked DjangoCMS page.
        """
        return self.filter(extended_object__publisher_is_draft=False)


class PageExtensionManager(models.Manager):
    """
    Add custom filters to the default Django manager for page extensions.
    """

    def get_queryset(self):
        """
        Use our custom queryset for page extensions.
        """
        return PageExtensionQuerySet(self.model, using=self._db)

    def drafts(self):
        """
        Make our custom filter "drafts" available on the page extension manager.
        """
        return self.get_queryset().drafts()

    def published(self):
        """
        Make our custom filter "published" available on the page extension manager.
        """
        return self.get_queryset().published()


class BasePageExtension(PageExtension):
    """
    The organization page extension represents and records entities that manage courses.
    It could be a university or a training company for example.

    This model should be used to record structured data about the organization whereas the
    associated page object is where we record the less structured information to display on the
    page to present the organization.
    """

    objects = PageExtensionManager()

    class Meta:
        abstract = True

    def check_publication(self, language=None):
        """
        Allow checking if a page extension is published without passing any language (unlike the
        "is_published" method on the page object): if not explicitly passed as argument, the
        language is retrieved from the context.

        The actual check is subcontracted to the "is_published" method on the related Django CMS
        Page object.

        Note: We choose not to name our method "is_published" like Django CMS, because it is a
        bad practice. Indeed, someone may think it is a property and use it without invocating it
        and the returned value (a bound method) will always be truthy... This issue happened a lot
        with the "is_authenticated" method on Django's User model before they converted it to a
        property.
        """
        language = language or translation.get_language()
        return self.extended_object.is_published(language)


class PagePluginMixin:
    """
    A base plugin model to build all our plugins that are used to publish a glimpse of a page on
    another page.
    """

    @property
    def relevant_page(self):
        """
        the plugin should show the published page whenever it exists or the draft page otherwise.

        On a public content, the draft page should not be shown at all but this is left to the
        caller.
        """
        if self.page.is_published(translation.get_language()):
            return self.page.get_public_object()
        return self.page

    def check_publication(self, language=None):
        """
        Allow checking if the page is published without passing any language (unlike the
        "is_published" method on the page object): if not explicitly passed as argument, the
        language is retrieved from the context.

        The actual check is subcontracted to the "is_published" method on the related Django CMS
        Page object.

        Note: We choose not to name our method "is_published" like Django CMS, because it is a
        bad practice. Indeed, someone may think it is a property and use it without invocating it
        and the returned value (a bound method) will always be truthy... This issue happened a lot
        with the "is_authenticated" method on Django's User model before they converted it to a
        property.
        """
        language = language or translation.get_language()
        return self.page.is_published(language)



"""Create_demo_site management command."""
import logging
import random
from collections import defaultdict

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand, CommandError
from django.test.utils import override_settings

import factory
from cms.api import add_plugin

from richie.apps.core.factories import create_text_plugin, image_getter
from richie.apps.core.helpers import recursive_page_creation
from richie.apps.courses.factories import (
    VIDEO_SAMPLE_LINKS,
    BlogPostFactory,
    CourseFactory,
    CourseRunFactory,
    LicenceFactory,
    OrganizationFactory,
    PersonFactory,
)

from ...defaults import (
    HOMEPAGE_CONTENT,
    ICONS_INFO,
    LEVELS_INFO,
    NB_OBJECTS,
    PAGES_INFO,
    SINGLECOLUMN_CONTENT,
    SITEMAP_PAGE_PARAMS,
    SUBJECTS_INFO,
)
from ...helpers import create_categories
from ...utils import pick_image

logger = logging.getLogger("richie.commands.demo.create_demo_site")


def get_number_of_course_runs():
    """
    Returns a random integer between 1 and the max number of course runs.
    We make it a convenience method so that it can be mocked in tests.
    """
    return random.randint(1, NB_OBJECTS["course_courseruns"])  # nosec


def get_number_of_icons():
    """
    Returns a random integer between 0 and the max number of course icons.
    0 is weighted to have as much chance to be chosen as all other numbers cumulated because
    we visually don't want all our courses to have an icon.
    We make it a convenience method so that it can be mocked in tests.
    """
    return random.choice(  # nosec
        [0] * NB_OBJECTS["course_icons"] + list(range(NB_OBJECTS["course_icons"]))
    )


# pylint: disable=no-member
#
# Looks like pylint is not relevant at guessing object types when cascading
# methods over querysets: Instance of 'list' has no 'delete' member (no-member).
# We choose to ignore this false positive warning.

# pylint: disable=too-many-locals,too-many-statements,too-many-branches
@override_settings(RICHIE_KEEP_SEARCH_UPDATED=False)
def create_demo_site():
    """
    Create a simple site tree structure for developpers to work in realistic environment.

    We create multilingual pages, add organizations under the related page and add
    plugins to each page.
    """
    site = Site.objects.get(id=1)

    # Create pages as described in PAGES_INFOS
    pages_created = recursive_page_creation(site, PAGES_INFO)

    # Create some licences
    licences = LicenceFactory.create_batch(
        NB_OBJECTS["licences"], logo__file__from_path=pick_image("licence")()
    )

    # Create organizations under the `Organizations` page
    organizations = OrganizationFactory.create_batch(
        NB_OBJECTS["organizations"],
        page_in_navigation=True,
        page_languages=["en", "fr"],
        page_parent=pages_created["organizations"],
        fill_banner=pick_image("banner"),
        fill_description=True,
        fill_logo=pick_image("logo"),
        should_publish=True,
        with_permissions=True,
    )

    # Generate each category tree and return a list of the leaf categories
    icons = list(
        create_categories(
            **ICONS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("logo"),
            page_parent=pages_created["categories"],
        )
    )
    levels = list(
        create_categories(
            **LEVELS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("logo"),
            page_parent=pages_created["categories"],
        )
    )
    subjects = list(
        create_categories(
            **SUBJECTS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("logo"),
            page_parent=pages_created["categories"],
        )
    )

    # Create persons under the `persons` page
    persons = []
    persons_for_organization = defaultdict(list)
    for _ in range(NB_OBJECTS["persons"]):
        # Randomly assign each person to a set of organizations
        person_organizations = random.sample(
            organizations,
            random.randint(1, NB_OBJECTS["person_organizations"]),  # nosec
        )
        person = PersonFactory(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["persons"],
            fill_categories=random.sample(
                subjects, random.randint(1, NB_OBJECTS["person_subjects"])  # nosec
            ),
            fill_organizations=person_organizations,
            fill_portrait=pick_image("portrait"),
            fill_bio=True,
            should_publish=True,
        )
        persons.append(person)
        for organization in person_organizations:
            persons_for_organization[organization.id].append(person)

    # Assign each person randomly to an organization so that our course are tagged realistically
    # If organizations and persons are tagged randomly on courses, each organizations will
    # in the end be related to most persons... not what we want.

    # Create courses under the `Course` page with categories and organizations
    # relations
    courses = []
    for _ in range(NB_OBJECTS["courses"]):
        video_sample = random.choice(VIDEO_SAMPLE_LINKS)  # nosec

        # Randomly assign each course to a set of organizations
        course_organizations = random.sample(
            organizations, NB_OBJECTS["course_organizations"]
        )

        # Only the persons members of these organizations are eligible to be part
        # of the course team
        eligible_persons = set(
            person
            for o in course_organizations
            for person in persons_for_organization[o.id]
        )

        course = CourseFactory(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["courses"],
            fill_licences=[
                ("course_license_content", random.choice(licences)),  # nosec
                ("course_license_participation", random.choice(licences)),  # nosec
            ],
            fill_team=random.sample(
                eligible_persons,
                min(
                    random.randint(1, NB_OBJECTS["course_persons"]),  # nosec
                    len(eligible_persons),
                ),
            ),
            fill_teaser=video_sample,
            fill_cover=pick_image("cover")(video_sample.image),
            fill_categories=[
                *random.sample(
                    subjects, random.randint(1, NB_OBJECTS["course_subjects"])  # nosec
                ),
                random.choice(levels),  # nosec
            ],
            fill_icons=random.sample(icons, get_number_of_icons()),
            fill_organizations=course_organizations,
            fill_texts=[
                "course_description",
                "course_format",
                "course_prerequisites",
                "course_plan",
                # "course_license_content",
                # "course_license_participation",
            ],
            should_publish=True,
        )
        course.create_permissions_for_organization(course_organizations[0])
        courses.append(course)

        # Add a random number of course runs to the course
        nb_course_runs = get_number_of_course_runs()
        # pick a subset of languages for this course (otherwise all courses will have more or
        # less all the languages across their course runs!)
        languages_subset = random.sample(
            ["de", "en", "es", "fr", "it", "nl"], random.randint(1, 4)  # nosec
        )
        for i in range(nb_course_runs):
            CourseRunFactory(
                __sequence=i,
                languages=random.sample(
                    languages_subset, random.randint(1, len(languages_subset))  # nosec
                ),
                page_in_navigation=False,
                page_languages=["en", "fr"],
                page_parent=course.extended_object,
                should_publish=True,
            )

    # Create blog posts under the `News` page
    blogposts = []
    for _ in range(NB_OBJECTS["blogposts"]):
        post = BlogPostFactory.create(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["blogposts"],
            fill_cover=pick_image("cover"),
            fill_excerpt=True,
            fill_body=True,
            fill_categories=[
                *random.sample(subjects, NB_OBJECTS["blogpost_categories"]),
                random.choice(levels),  # nosec
            ],
            fill_author=random.sample(persons, 1),
            should_publish=True,
        )
        blogposts.append(post)

    # Once everything has been created, use some content to create a homepage
    placeholder = pages_created["home"].placeholders.get(slot="maincontent")

    # - Get a banner image
    banner = image_getter(pick_image("banner")())

    # - Get a logo image
    logo = image_getter(pick_image("logo")())

    # - Create the home page in each language
    for language, content in HOMEPAGE_CONTENT.items():
        # Add a banner
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LargeBannerPlugin",
            title=content["banner_title"],
            background_image=banner,
            logo=logo,
            logo_alt_text="logo",
            content=content["banner_content"],
            template=content["banner_template"],
        )
        # Add highlighted courses with a button
        courses_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["courses_title"],
            template=content["section_template"],
        )
        for course in random.sample(courses, NB_OBJECTS["home_courses"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CoursePlugin",
                target=courses_section,
                page=course.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=courses_section,
            name=content["courses_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["courses"],
        )

        # Add highlighted blogposts
        blogposts_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["blogposts_title"],
            template=content["section_template"],
        )
        for blogpost in random.sample(blogposts, NB_OBJECTS["home_blogposts"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="BlogPostPlugin",
                target=blogposts_section,
                page=blogpost.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=blogposts_section,
            name=content["blogposts_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["blogposts"],
        )

        # Add highlighted organizations
        organizations_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["organizations_title"],
            template=content["section_template"],
        )
        for organization in random.sample(
            organizations, NB_OBJECTS["home_organizations"]
        ):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="OrganizationPlugin",
                target=organizations_section,
                page=organization.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=organizations_section,
            name=content["organizations_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["organizations"],
        )

        # Add highlighted subjects
        subjects_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["subjects_title"],
            template=content["section_template"],
        )
        for subject in random.sample(subjects, NB_OBJECTS["home_subjects"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CategoryPlugin",
                target=subjects_section,
                page=subject.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=subjects_section,
            name=content["subjects_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["categories"],
        )

        # Add highlighted persons
        persons_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["persons_title"],
            template=content["section_template"],
        )
        for person in random.sample(persons, NB_OBJECTS["home_persons"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="PersonPlugin",
                target=persons_section,
                page=person.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=persons_section,
            name=content["persons_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["persons"],
        )

        # Once content has been added we must publish again homepage
        pages_created["home"].publish(language)

    # Fill the single column sample page
    placeholder = pages_created["annex__about"].placeholders.get(slot="maincontent")

    # - Get a banner image
    banner = image_getter(pick_image("banner")())

    # - Get a logo image
    logo = image_getter(pick_image("logo")())

    # - Get a video
    video_sample = random.choice(VIDEO_SAMPLE_LINKS)  # nosec

    # - Create sample page in each language
    for language, content in SINGLECOLUMN_CONTENT.items():
        # Add a banner
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LargeBannerPlugin",
            title=content["banner_title"],
            background_image=banner,
            content=content["banner_content"],
            template=content["banner_template"],
        )
        # HTML paragraphs
        create_text_plugin(
            pages_created["annex__about"],
            placeholder,
            nb_paragraphs=random.randint(3, 4),  # nosec
            languages=[language],
            plugin_type="TextPlugin",
        )
        # A large video sample
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="VideoPlayerPlugin",
            label=video_sample.label,
            embed_link=video_sample.url,
            template="full-width",
        )
        # Section with some various plugins
        sample_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["section_sample_title"],
            template=content["section_sample_template"],
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="OrganizationPlugin",
            target=sample_section,
            page=random.choice(organizations).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="CoursePlugin",
            target=sample_section,
            page=random.choice(courses).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="OrganizationPlugin",
            target=sample_section,
            page=random.choice(organizations).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="BlogPostPlugin",
            target=sample_section,
            page=random.choice(blogposts).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=sample_section,
            name=content["section_sample_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["home"],
        )
        # Add a licence
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LicencePlugin",
            licence=random.choice(licences),  # nosec
        )
        # Add a simple picture entry
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SimplePicturePlugin",
            picture=logo,
        )
        # Add a plain text
        text = factory.Faker(
            "text", max_nb_chars=random.randint(150, 250), locale=language  # nosec
        ).generate({})
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="PlainTextPlugin",
            body=text,
        )

        # Once content has been added we must publish again the about page
        pages_created["annex__about"].publish(language)

    # Create a sitemap page
    placeholder = pages_created["annex__sitemap"].placeholders.get(slot="maincontent")

    for language in pages_created["annex__sitemap"].get_languages():
        parent_instance = add_plugin(
            language=language, placeholder=placeholder, plugin_type="HTMLSitemapPlugin"
        )
        for name, params in SITEMAP_PAGE_PARAMS.items():
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="HTMLSitemapPagePlugin",
                target=parent_instance,
                root_page=pages_created[name],
                **params,
            )

        # Once content has been added we must publish again the sitemap
        pages_created["annex__sitemap"].publish(language)


class Command(BaseCommand):
    """Create default pages for FUN frontend"""

    help = __doc__

    def add_arguments(self, parser):

        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="Force command execution despite DEBUG is set to False",
        )

    def handle(self, *args, **options):

        if not settings.DEBUG and not options["force"]:
            raise CommandError(
                (
                    "This command is not meant to be used in production environment "
                    "except you know what you are doing, if so use --force parameter"
                )
            )

        create_demo_site()

        logger.info("done")



"""
Courses CMS plugins
"""

from django.utils.translation import ugettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from richie.apps.core.defaults import PLUGINS_GROUP

from .forms import LicencePluginForm
from .models import (
    BlogPostPluginModel,
    CategoryPluginModel,
    CoursePluginModel,
    LicencePluginModel,
    OrganizationPluginModel,
    PersonPluginModel,
)


@plugin_pool.register_plugin
class OrganizationPlugin(CMSPluginBase):
    """
    Organization plugin displays an organization's information on other pages
    """

    cache = True
    model = OrganizationPluginModel
    module = PLUGINS_GROUP
    render_template = "courses/plugins/organization.html"

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context


@plugin_pool.register_plugin
class CategoryPlugin(CMSPluginBase):
    """
    Category plugin displays a category's information on other pages
    """

    cache = True
    model = CategoryPluginModel
    module = PLUGINS_GROUP
    render_template = "courses/plugins/category_plugin.html"

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context


@plugin_pool.register_plugin
class CoursePlugin(CMSPluginBase):
    """
    Course plugin displays a course's information on other pages
    """

    cache = True
    model = CoursePluginModel
    module = PLUGINS_GROUP
    render_template = "courses/plugins/course_plugin.html"

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context


@plugin_pool.register_plugin
class PersonPlugin(CMSPluginBase):
    """
    Person plugin displays a person's information on other pages
    """

    cache = True
    model = PersonPluginModel
    module = PLUGINS_GROUP
    render_template = "courses/plugins/person.html"

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context


@plugin_pool.register_plugin
class LicencePlugin(CMSPluginBase):
    """
    CMSPlugin to add a licence selected from available Licence objects and
    with an additional free text.
    """

    allow_children = False
    cache = True
    fieldsets = ((None, {"fields": ["licence", "description"]}),)
    form = LicencePluginForm
    model = LicencePluginModel
    module = PLUGINS_GROUP
    name = _("Licence")
    render_template = "courses/plugins/licence_plugin.html"

    def render(self, context, instance, placeholder):
        context.update({"instance": instance, "placeholder": placeholder})
        return context


@plugin_pool.register_plugin
class BlogPostPlugin(CMSPluginBase):
    """
    BlogPost plugin displays a Blog post overview on other pages
    """

    cache = True
    model = BlogPostPluginModel
    module = PLUGINS_GROUP
    render_template = "courses/plugins/blogpost.html"

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "relevant_page": instance.relevant_page,
                "placeholder": placeholder,
            }
        )
        return context



"""
Simple text plugin forms
"""
from django import forms

from djangocms_text_ckeditor.widgets import TextEditorWidget

from .models import SimpleText


class CKEditorPluginForm(forms.ModelForm):
    """
    Plugin form used to fill its content from frontend admin.
    """

    class Meta:
        """
        Form meta attributes
        """

        model = SimpleText
        widgets = {"body": TextEditorWidget}
        fields = ["body"]



"""
Helpers that can be useful throughout the whole project
"""
from functools import reduce
from operator import or_

from django.conf import settings
from django.contrib.auth.models import Permission
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.utils.text import slugify

from cms.api import create_page, create_title
from cms.models import Page


def get_permissions(names):
    """
    Given an iterable of permission names of the form "app_label.codename",
    return an iterable of the corresponding existing Permission objects.
    """
    names = set(names)  # Eliminate redundancies
    split_names = (name.split(".", 1) for name in names)
    query_elements = [
        Q(content_type__app_label=app_label, codename=codename)
        for app_label, codename in split_names
    ]
    permissions = (
        Permission.objects.filter(reduce(or_, query_elements))
        if query_elements
        else Permission.objects.none()
    )

    if len(permissions) != len(names):
        differences = names - {
            f"{p.content_type.app_label:s}.{p.codename:s}" for p in permissions
        }
        raise Permission.DoesNotExist(
            "Some permission names were not found: {:s}".format(", ".join(differences))
        )

    return permissions


def create_i18n_page(title, languages=None, is_homepage=False, **kwargs):
    """
    Creating a multilingual page is not straightforward so we should have a helper

    This content argument should be a dictionary with the title of the page in each language:

        {
            'en': 'About',
            'fr': 'A propos',
            'de': 'Impressum',
        }

    """
    kwargs["template"] = kwargs.get("template", "richie/single_column.html")

    if isinstance(title, dict):
        # Check that the languages passed are coherent with the languages requested if any
        if languages:
            invalid_languages = set(languages) - set(title.keys())
            if invalid_languages:
                raise ValueError(
                    "Page titles are missing in some requested languages: {:s}".format(
                        ",".join(invalid_languages)
                    )
                )
        else:
            languages = title.keys()
        i18n_titles = title

    elif isinstance(title, str):
        # Add a marker at the end of the string to differentiate each language
        languages = languages or [settings.LANGUAGE_CODE]
        i18n_titles = {language: title for language in languages}

    else:
        raise ValueError(
            "Title should be a string or a dictionary of language/string pairs"
        )

    # Assert that the languages passed are declared in settings
    invalid_languages = set(languages) - {l[0] for l in settings.LANGUAGES}
    if invalid_languages:
        raise ValueError(
            "You can't create pages in languages that are not declared: {:s}".format(
                ",".join(invalid_languages)
            )
        )

    # Make a copy of languages to avoid muting it in what follows
    languages = list(languages)
    # Create the page with a first language from what is given to us
    first_language = languages.pop(0)

    slug = slugify(i18n_titles[first_language])
    page = create_page(
        language=first_language, title=i18n_titles[first_language], slug=slug, **kwargs
    )

    if is_homepage is True:
        page.set_as_homepage()

    # Add a title for each additional language
    for language in languages:
        create_title(
            language=language,
            title=i18n_titles[language],
            slug=slugify(i18n_titles[language]),
            page=page,
        )
        # Publish page in each additional language
        if kwargs.get("published") is True:
            page.publish(language)

    return page


def recursive_page_creation(site, pages_info, parent=None):
    """
    Recursively create page following tree structure with parent/children.

    Arguments:
        site (django.contrib.sites.models.Site): Site object which page will
            be linked to.
        pages_info (dict): Page items to create recursively such as 'children' key
            value can be a dict to create child pages. The current page is
            given to children for parent relation.

    Keyword Arguments:
        parent (cms.models.pagemodel.Page): Page used as a parent to create
            page item from `pages` argument.

    Returns:
        dict: mapping of the page names passed in argument and the created page instances.
    """
    pages = {}

    for name, kwargs in pages_info.items():
        children = kwargs.pop("children", None)

        if kwargs.get("is_homepage"):
            query_params = {"is_home": True}
        else:
            query_params = {"reverse_id": name}

        try:
            page = Page.objects.get(
                publisher_is_draft=True, node__site=site, **query_params
            )
        except Page.DoesNotExist:
            page = create_i18n_page(
                site=site, parent=parent, published=True, reverse_id=name, **kwargs
            )

        pages[name] = page

        # Create children
        if children:
            children_pages = recursive_page_creation(site, children, parent=page)
            for child_name in children_pages:
                if child_name in pages:
                    raise ImproperlyConfigured(
                        "Page names should be unique: {:s}".format(child_name)
                    )
            pages.update(children_pages)

    return pages



"""
Handle ElasticSearch setup that needs to be done at bootstrap time.
"""
import logging

from django.core.management.base import BaseCommand

from ...index_manager import regenerate_indexes, store_es_scripts

logger = logging.getLogger("richie.search.bootstrap_elasticsearch")


class Command(BaseCommand):
    """
    Bootstrap tasks include:
    - create indices for courses, organizations, categories,
    - index all records in their respective indices,
    - store necessary scripts.
    """

    help = __doc__

    def handle(self, *args, **options):
        # Keep track of starting time for logging purposes
        logger.info("Starting to regenerate ES indices...")

        # Creates new indices each time, populates them, and atomically replaces
        # the old indices once the new ones are ready.
        regenerate_indexes(logger)

        # Confirm operation success through a console log
        logger.info("ES indices regenerated.")

        logger.info("Starting to store ES scripts...")

        # Iterates over indexables to find all the necessary scripts and store them
        store_es_scripts(logger)

        logger.info("ES scripts stored.")



"""
A custom field that defines efforts as a number of pre-defined time units per reference time unit
(e.g 2 hours/day, 5 hours/week or 3 days/months).
"""
from django import forms
from django.core import checks, exceptions
from django.db import models
from django.db.utils import DataError
from django.forms import widgets
from django.utils.translation import ugettext_lazy as _


class EffortWidget(widgets.MultiWidget):
    """
    A widget that splits the input of an effort into one integer input box and two select boxes.
    """

    template_name = "richie/forms/widgets/composite_widget.html"

    def __init__(
        self,
        attrs=None,
        choices=(),
        default_effort_unit=None,
        default_reference_unit=None,
    ):
        """
        Split the field in 3 widgets:
        - the first widget is a positive integer input,
        - the second widget is a select box to choose a pre-defined time unit (minutes, hours,
          days, weeks or months),
        - the third widget is a select box to choose the pre-defined time unit of reference.

        e.g: 3 hours/day is split in: 3 (integer input) | hour (select) | day (select)
        """
        self.default_effort_unit = default_effort_unit
        self.default_reference_unit = default_reference_unit
        super().__init__(
            (
                widgets.NumberInput({**(attrs or {}), "min": 0}),
                # Remove the last choice: it can never be chosen as it must be strictly smaller
                # than the reference time unit
                widgets.Select(attrs, choices[:-1]),
                # Remove the first choice: it can never be chosen as it must be strictly greater
                # than the effort time unit
                widgets.Select(attrs, choices[1:]),
            )
        )

    def decompress(self, value):
        """
        Return the raw value as it is already a triplet [duration, unit, reference].
        In the absence of a value, pre-configure the two time unit select widgets as defined
        in the field.
        """
        return value or ["", self.default_effort_unit, self.default_reference_unit]


class EffortFormField(forms.CharField):
    """
    The composite duration field is saved in database as a charfield but represented in Python
    by a triplet [duration,unit,reference]. We should not force its value to str.
    """

    def to_python(self, value):
        """Return the raw value to avoid the charfield behavior that forces the value to str."""
        return value


class EffortField(models.CharField):
    """
    A custom charfield to store an effort in database as an integer and 2 time unit choices.
    """

    description = _("Define an effort")

    error_message = _("%(value)s is not a valid choice for a time unit.")

    def __init__(self, *args, **kwargs):
        """
        Record the three additional field attributes `time_units`, `default_effort_unit` and
        `default_reference_unit` in the instance.
        """
        self.time_units = kwargs.pop("time_units", None)
        self.default_effort_unit = kwargs.pop("default_effort_unit", None)
        self.default_reference_unit = kwargs.pop("default_reference_unit", None)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """Return enough information to recreate the field as a 4-tuple."""
        name, path, args, kwargs = super().deconstruct()
        kwargs["time_units"] = self.time_units
        kwargs["default_effort_unit"] = self.default_effort_unit
        kwargs["default_reference_unit"] = self.default_reference_unit
        return name, path, args, kwargs

    def _check_time_units_attribute(self):
        """Check that `time_units` is well configured."""
        if self.time_units is None:
            return [
                checks.Error(
                    "Effort fields must define a 'time_units' attribute.",
                    obj=self,
                    id="fields.E1010",
                )
            ]
        if not (
            isinstance(self.time_units, dict)
            and all(
                [isinstance(c, tuple) and len(c) == 2 for c in self.time_units.values()]
            )
        ):
            return [
                checks.Error(
                    "'time_units' must be a dictionary of tuples.",
                    obj=self,
                    id="fields.E1011",
                )
            ]

        return []

    def _check_default_effort_unit_attribute(self):
        """Check that `default_effort_unit` is well configured."""
        if self.default_effort_unit and self.default_effort_unit not in self.time_units:
            return [
                checks.Error(
                    "'{:s}' is not a valid time unit.".format(self.default_effort_unit),
                    obj=self,
                    id="fields.E1012",
                )
            ]

        return []

    def _check_default_reference_unit_attribute(self):
        """Check that `default_reference_unit` is well configured."""
        if (
            self.default_reference_unit
            and self.default_reference_unit not in self.time_units
        ):
            return [
                checks.Error(
                    "'{:s}' is not a valid time unit.".format(
                        self.default_reference_unit
                    ),
                    obj=self,
                    id="fields.E1013",
                )
            ]

        return []

    def check(self, **kwargs):
        """Add checks on time units to the field's checks."""
        return (
            super().check(**kwargs)
            + self._check_time_units_attribute()
            + self._check_default_effort_unit_attribute()
            + self._check_default_reference_unit_attribute()
        )

    @staticmethod
    def from_db_value(value, *_args):
        """Convert a database value to a list value."""
        if not value:
            return None

        try:
            duration, unit, reference = value.split("|")
        except ValueError:
            raise DataError(
                "Value in database should be of the form '{duration:d}|{unit:s}|{reference:s}'"
            )

        if not duration:
            return None

        return int(duration), unit, reference

    def to_python(self, value):
        """Convert a string value to a list value. Used for deserialization and in clean forms."""
        if isinstance(value, (list, tuple)):
            return value

        if not value:
            return None

        if isinstance(value, str):
            try:
                duration, unit, reference = value.split("|")
            except ValueError:
                raise DataError(
                    "Value in database should be of the form '{duration:d}|{unit:s}|{reference:s}'"
                )

            return int(duration.strip()), unit.strip(), reference.strip()

        return value

    def get_prep_value(self, value):
        """
        Transform the list value to a concatenation of "|" separated strings.

        Arguments:
        ----------
        value (List[string]): a list of strings representing the Python representation
            of an effort.

        Returns:
        --------
        string:
            None > None
            [0, "day", "week"] > "0|day|week"
            [5, "day", "week"] > "5|day|week"

        """
        if value is None:
            return None
        return "|".join([str(v).strip() for v in value])

    def value_to_string(self, obj):
        """Serialize the Python value. We can use existing methods as it is a CharField."""
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def validate(self, value, model_instance):
        """
        Validate each value in values and raise a ValidationError if something is wrong.
        """
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value:
            try:
                size = len(value)
            except TypeError:
                size = 0

            if size != 3:
                raise exceptions.ValidationError(
                    _(
                        "An effort should be a triplet: number, time unit and reference unit."
                    ),
                    code="invalid_format",
                )

            duration, unit, reference = value

            if duration:
                # Check that the duration is an integer
                try:
                    duration = int(duration)
                except ValueError:
                    raise exceptions.ValidationError(
                        _("An effort should be a round number of time units."),
                        code="invalid_effort",
                    )

                # Check that the duration is positive
                if duration <= 0:
                    raise exceptions.ValidationError(
                        _("An effort should be positive."), code="negative_effort"
                    )

            choices = list(dict(self.time_units))

            # Search the time unit in this set
            if unit not in choices:
                raise exceptions.ValidationError(
                    self.error_message, code="invalid_interface", params={"value": unit}
                )

            # Search the reference time unit in this set
            if reference not in choices:
                raise exceptions.ValidationError(
                    self.error_message,
                    code="invalid_reference",
                    params={"value": reference},
                )

            if choices.index(unit) >= choices.index(reference):
                raise exceptions.ValidationError(
                    _(
                        "The effort time unit should be shorter than the reference unit."
                    ),
                    code="unit_order",
                )

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages["null"], code="null")

        if not self.blank and not value:
            raise exceptions.ValidationError(self.error_messages["blank"], code="blank")

    def formfield(self, **kwargs):
        """Declare the above EffortFormField as form field for this model field."""
        kwargs = {
            **kwargs,
            "form_class": EffortFormField,
            "widget": EffortWidget(
                choices=tuple(
                    (name, strings[0]) for name, strings in self.time_units.items()
                ),
                default_effort_unit=self.default_effort_unit,
                default_reference_unit=self.default_reference_unit,
            ),
        }
        return super().formfield(**kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        """
        Add a method to the model using this field to enable the common `get_FIELD_display`
        pattern to display the human readable version of a field based on choices.

        We choose to display a sentence of the form: "5 days/week".
        """
        super().contribute_to_class(cls, name, private_only=private_only)
        choices = self.time_units

        def func(self, *_args):
            """
            we add `*args` because this method is used in a "render_model" template tag
            and it passes a "request" argument when calling the method.
            For more information, see:
            http://docs.django-cms.org/en/latest/how_to/frontend_models.html
            """
            value = getattr(self, name)
            if value:
                duration, unit, reference = value
                duration = int(duration)
                count_index = 1 if duration > 1 else 0
                return "{duration:d} {unit!s}/{reference!s}".format(
                    duration=duration,
                    unit=choices[unit][count_index],
                    reference=choices[reference][0],
                )
            return ""

        setattr(cls, "get_%s_display" % self.name, func)



"""Django Configuration mixins for the Richie courses app."""
from configurations.utils import uppercase_attributes

from .. import settings


class RichieCoursesConfigurationMixin:
    """
    A Django Configuration mixin to set sensible defaults for the settings of Richie's
    courses app.
    """

    def __new__(cls):
        """
        Set all settings as attributes of this class so it can be used as a mixin in
        Django Configuration.
        """
        for key, value in uppercase_attributes(settings).items():
            setattr(cls, key, value)
        return super().__new__(cls)



"""
Define base FilterDefinition classes on which to build actual FilterDefinition classes.
They encapsulate common behavior around managing filters in ElasticSearch queries and
API requests (to validate input) and responses (to produce easy-to-consume filters).
"""
from functools import reduce
from operator import itemgetter

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from ..defaults import FACET_SORTING_DEFAULT


# pylint: disable=too-many-instance-attributes
class BaseFilterDefinition:
    """
    Base filter definition shape: just a skeleton.
    """

    SORTING_CONF, SORTING_COUNT, SORTING_NAME = "conf", "count", "name"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        term=None,
        human_name=None,
        is_autocompletable=False,
        is_drilldown=False,
        min_doc_count=0,
        position=0,
        sorting=FACET_SORTING_DEFAULT,
    ):
        """Set common attributes with sensible defaults (only name is required)."""
        self.name = name
        self.term = term or name
        self.human_name = human_name or name
        self.is_autocompletable = is_autocompletable
        self.is_drilldown = is_drilldown
        self.min_doc_count = min_doc_count
        self.position = position
        self.sorting = sorting

    def get_form_fields(self):
        """
        Returns:
        --------
            django.forms.fields.Field: the form field instance that will be used to parse
                and validate this filter's values from the querystring.

        To be implemented in each actual FilterDefintion class.
        """
        raise NotImplementedError()

    def get_query_fragment(self, data):
        """
        Build the query fragment to use in the ElasticSearch filter & aggregations.

        Arguments:
        ----------
            data (Dict): a dictionary mapping the name of filters with the list of
                values selected for each filter. Typically the `cleaned_data` output of
                a valid filter form:
                e.g. {"availability": ["open"], "languages": ["en", "fr"]}

        Returns:
        --------
            List[Dict]: a list of dictionaries each mapping the name of a filter with a query
                fragment that must be added to the global Elasticsearch query to apply this
                filter. For example:
                [
                    {
                        "key": "new",
                        "fragment": [{'term': {'is_new': True}}]
                    }
                ]

        To be implemented in each actual FilterDefintion class.
        """
        raise NotImplementedError()

    def get_aggs_fragment(self, queries, *args, **kwargs):
        """
        Build the aggregations fragment to use to extract facets from ElasticSearch.

        Arguments:
        ----------
            queries (List[Dict]): a list of all the query fragments composing the global
                Elasticsearch query and which were collected by calling the `get_query_fragment`
                method of each filter definitions. Each item in the list is a dictionary mapping
                the name of a filter with a query fragment (see above for an example).

        Returns:
        --------
            Dict: a dictionary mapping the name of an aggregation bucket with an aggregation
                fragment that will be added to the global Elasticsearch aggregation query to get
                facet counts related to this filter. For example:
                {
                    # A term aggregation query
                    "categories": {
                        "aggregations": {
                            "categories": {"terms": {"field": "categories"}}
                        },
                    },
                    # A filter aggregation query
                    "new@new": {
                        "filter": {
                            "bool": {
                                "must": [{"term": {"is_new": True}}]
                            }
                        }
                    }
                }

        To be implemented in each actual FilterDefintion class.
        """
        raise NotImplementedError()

    def get_faceted_definitions(self, facets, *args, **kwargs):
        """
        Build the filter definition from a filter's common attributes and its Elasticsearch
        facet results. The frontend uses these definitions to display the filters on the
        user interface (see search page left side bar).

        Arguments:
        ----------
            Dict: a dictionary mapping each aggregation name (one for each aggregation bucket
                defined by the `get_aggs_fragment` method) with its documents counts as returned
                by Elasticsearch in the "aggregations" part of the response.

        Returns:
        --------
            Dict: a dictionary mapping the name of a filter with a dictionary of all the
                information necessary to display the filter on the search page with its
                facet counts. For example:

                {
                    languages: {
                        "human_name": "Languages",
                        "is_drilldown": False,
                        "name": "languages",
                        "values": [
                            {"key": "en", "human_name": "English", "count": 3},
                            {"key": "fr", "human_name": "French", "count": 2},
                        ],
                    }
                }

        To be implemented in each actual FilterDefintion class.
        """
        raise NotImplementedError()


class BaseChoicesFilterDefinition(BaseFilterDefinition):
    """
    A base filter definition to filter on a field that has a pre-defined list of possible
    values. The values and their related query fragment must be defined in the actual filter
    class.
    """

    def get_values(self):
        """
        Get the possible values for this choice filter.

        Returns:
        --------
            Dict: a dictionary mapping the machine name of a value with its human name.
                For example: {"is_new": "First session", "is_repeat": "Repeat session"}
        """
        raise NotImplementedError()

    def get_fragment_map(self):
        """
        Get the possible query fragments for this choice filter.

        Returns:
        --------
            Dict: a dictionary mapping the machine name of a value with its human name.
                For example:
                {
                    "is_new": [{"term": {"is_new": True}}],
                    "is_repeat": [{"term": {"is_new": False}}],
                }
        """
        raise NotImplementedError()

    def get_form_fields(self):
        """Choice filters are validated with a MultipleChoiceField."""
        return {
            self.name: (
                forms.MultipleChoiceField(
                    required=False, choices=self.get_values().items()
                ),
                True,  # a MultipleChoiceField expects list values
            )
        }

    def get_faceted_definitions(self, facets, *args, **kwargs):
        """
        Add the counts to the values from the initial definition to make them complete
        and sorted as the frontend expects them.
        """
        # Get filter values to derive human names
        values = self.get_values()

        # for each facet, we derive the value and the count:
        # eg. for facet key `availability@coming_soon`, the value is `coming_soon`
        facet_counts_dict = {
            key.split("@")[1]: facet["doc_count"]
            for key, facet in facets.items()
            if "{:s}@".format(self.name) in key  # 6 times faster than startswith
            and facet["doc_count"] >= self.min_doc_count
        }

        # Sort facets as requested
        if self.sorting == self.SORTING_NAME:
            # Alphabetical ascending sorting
            facet_counts = sorted(facet_counts_dict.items(), key=itemgetter(0))
        elif self.sorting == self.SORTING_COUNT:
            # Sorting by descending facet count
            facet_counts = sorted(
                facet_counts_dict.items(), key=itemgetter(1), reverse=True
            )
        elif self.sorting == self.SORTING_CONF:
            # Respect the order set in filter definitions
            facet_counts = [
                (name, facet_counts_dict[name])
                for name in values
                if name in facet_counts_dict
            ]
        else:
            raise ImproperlyConfigured(
                'Facet sorting should be one of "conf", "count" or "name"'
            )

        return {
            self.name: {
                # We always need to pass the base definition to the frontend
                "human_name": self.human_name,
                "is_autocompletable": self.is_autocompletable,
                "is_drilldown": self.is_drilldown,
                "name": self.name,
                "position": self.position,
                "values": [
                    {"count": count, "human_name": values[name], "key": name}
                    for name, count in facet_counts
                ],
            }
        }


class NestingWrapper(BaseFilterDefinition):
    """
    A filter definition to aggregate children filter definitions under a nested query.

    This is necessary to use filters and aggregations on values that live in nested fields.
    Such as any information that is linked not directly to courses but to course runs (a nested
    field of courses).
    """

    def __init__(self, name, filters, path=None):
        """
        Instanciate all the children filter definitions and record them as an attribute
        in order to wrap calls to all the common methods:
            - get_form_fields,
            - get_query_fragment,
            - get_aggs_fragment,
            - get_faceted_definitions.

        Arguments:
        ----------
            name (string): the name of the nesting wrapper (not really a filter definition as it
                just collects queries/aggs from  its children.
            filters (List[Tuple]): list of children filters to be instantiated, each defined by
                a tuple of 2 elements:
                - class: dotted path pointing to the filter definition class to be instiated,
                - kwargs: dictionary of the keyword arguments to instantiate the children filter
                    definitions.
            path (string): the name of the nested field under which children filters will query.
                optional as it defaults to the value passed for the `name` argument.
        """
        super().__init__(name)
        self.path = path or name
        self.filter_definitions = {
            params["name"]: import_string(dotted_path)(
                term="{:s}.{:s}".format(self.path, params["name"]), **params
            )
            for dotted_path, params in filters
        }

    def get_form_fields(self):
        """
        Aggregate form fields from the nested children in a mapping.

        Returns:
        --------
            Dict: dictionary mapping the key of each children filter definition with its
                formfield. The form thus gets at the same level the fields for nested filters
                and normal filters.

                As a result, filtering on a nested field is transparent in the querystring.
        """
        return reduce(
            lambda key_field_map, filter_definition: {
                **key_field_map,
                **filter_definition.get_form_fields(),
            },
            self.filter_definitions.values(),
            {},
        )

    def get_query_fragment(self, data):
        """
        Aggregate query fragments from the nested children in a list of mappings.

        Arguments:
        ----------
            data (Dict): a dictionary mapping the name of filters with the list of
                values selected for each filter. Typically the `cleaned_data` output of
                a valid filter form:
                e.g. {"availability": ["open"], "languages": ["en", "fr"]}

        Returns:
        --------
            List[Dict]: a list of one dictionary mapping the name of the nesting wrapper with a
                nested query cumulating all the active filters among the nested children.

        For example, if the children return query fragments as follows:
            - availability (archived):
                {
                    'key': 'availability',
                    'fragment': [
                        {
                            'bool': {
                                'must': [{'range': {'course_runs.end': {'lte': '2019-03-08'}}}]
                            }
                        }
                    ]
                }
            - languages (french or english):
                {
                    'key': 'languages',
                    'fragment': [
                        {
                            'bool': {
                                'must': [{'terms': {'course_runs.languages': ['en', 'fr']}}]
                            }
                        }
                    ]
                }

        Then the nesting wrapper should return a list with one aggregated nested query:
        [
            {
                'key': 'course_runs',
                'fragment': [
                    {
                        'nested': {
                            'path': 'course_runs',
                            'query': {
                                'bool': {
                                    'must': [
                                        {'range': {'course_runs.end': {'lte': '2019-03-08}}},
                                        {'terms': {'course_runs.languages': ['en', 'fr']}},
                                    ]
                                }
                            }
                        }
                    }
                ]
            }
        ]
        """
        queries = [
            kf_pair
            for fd in self.filter_definitions.values()
            for kf_pair in fd.get_query_fragment(data)
        ]
        nested_query = {
            "bool": {
                "must":
                # queries => map(pluck("fragment")) => flatten()
                [clause for kf_pair in queries for clause in kf_pair["fragment"]]
            }
        }
        return (
            [
                {
                    "key": self.name,
                    "fragment": [
                        {"nested": {"path": self.path, "query": nested_query}}
                    ],
                }
            ]
            if queries
            else []
        )

    # pylint: disable=arguments-differ
    def get_aggs_fragment(self, queries, data, *args, **kwargs):
        """
        Collect aggregations fragments from the nested children in a mapping.

        Arguments:
        ----------
            queries (List[Dict]): a list of all the query fragments composing the global
                Elasticsearch query and which were collected by calling the `get_query_fragment`
                method of each filter definitions. Each item in the list is a dictionary mapping
                the name of a filter with a query fragment (see above for an example).
            data (Dict): a dictionary mapping the name of filters with the list of
                values selected for each filter. Typically the `cleaned_data` output of
                a valid filter form:
                e.g. {"availability": ["open"], "languages": ["en", "fr"]}

        Returns:
        --------
            List[Dict]:a dictionary mapping the name of aggregation buckets for all the children
                the nesting wrapper, with an aggregation fragment that will be added to the
                global Elasticsearch aggregation query to get facet counts related to this value.

                We can't use term query to face on nested fields because what want to count is the
                number of courses that are not excluded by the applied filters, not the number of
                occurence of a value on a nested field that may be repeated under a given course!
                We can only use "filter" aggregations based on a nested query that excludes the
                current value on which we are facetting.

                For example, when filering on archived courses in french or english, the nested
                query is:
                [
                    {
                        'key': 'course_runs',
                        'fragment': [
                            {
                                'nested': {
                                    'path': 'course_runs',
                                    'query': {
                                        'bool': {
                                            'must': [
                                                {'range': {
                                                    'course_runs.end': {'lte': '2019-03-08'}}
                                                },
                                                {'terms': {'course_runs.languages': ['en', 'fr']}},
                                            ]
                                        }
                                    }
                                }
                            }
                        ]
                    }
                ]

                The aggregations buckets on values of the nested fields should look as follows.

                - To count on-going courses respecting this query:
                'availability@ongoing': {
                    'filter': {'bool': {'must': [
                        {'nested': {
                            'path': 'course_runs',
                            'query': {'bool': {'must': [
                                {'range': {'course_runs.start': {'lte': '2019-03-08'}}},
                                {'range': {'course_runs.end': {'gte': '2019-03-08}}},
                                {'terms': {'course_runs.languages': ['en']}},
                            ]}},
                        }},
                    ]}},
                }

                - To count french courses respecting this query:
                'languages@de': {
                    'filter': {'bool': {'must': [
                        {'nested': {
                            'path': 'course_runs',
                            'query': {'bool': {'must': [
                                {'range': {'course_runs.end': {'lte': '2019-03-08'}}},
                                {'terms': {'course_runs.languages': ['fr']}},
                            ]}},
                        }},
                    ]}},
                }

                These queries seem very difficult to build but luckily, the nesting wrapper
                (parent) knows how to build this nested queries via its `get_query_fragment`
                method. So we start by removing the nested query (as calculated with active
                filters) from the list of active queries and pass the parent as argument to
                each children so the children can rebuild the query for each of the value,
                excluding values one-by-one.
        """
        # Remove the query fragment built by this nesting wrapper. The children will call the
        # parent's `get_query_fragment` with specific data to build aggregation queryies one-
        # by-one.
        stripped_queries = list(
            filter(lambda query_fragment: query_fragment["key"] != self.name, queries)
        )

        aggs = {}
        for filter_definition in self.filter_definitions.values():
            aggs.update(
                filter_definition.get_aggs_fragment(
                    stripped_queries, data, parent=self, *args, **kwargs
                )
            )
        return aggs

    def get_faceted_definitions(self, facets, *args, **kwargs):
        """
        Collect facet definitions from the children filter definitions in a dictionary.
        """
        return {
            name: facet_definition
            for fd in self.filter_definitions.values()
            for name, facet_definition in fd.get_faceted_definitions(
                facets, *args, **kwargs
            ).items()
        }



"""
ElasticSearch indexes utilities.
"""
from functools import reduce

from django.conf import settings
from django.utils import timezone

from elasticsearch.client import IndicesClient
from elasticsearch.exceptions import NotFoundError
from elasticsearch.helpers import bulk

from . import ES_CLIENT
from .defaults import ES_CHUNK_SIZE
from .indexers import ES_INDICES
from .text_indexing import ANALYSIS_SETTINGS


def richie_bulk(actions):
    """Wrap bulk helper to set default parameters."""
    return bulk(
        actions=actions,
        chunk_size=getattr(settings, "RICHIE_ES_CHUNK_SIZE", ES_CHUNK_SIZE),
        client=ES_CLIENT,
        stats_only=True,
    )


def get_indexes_by_alias(existing_indexes, alias):
    """
    Get existing index(es) for an alias. Support multiple existing aliases so the command
    always 'cleans up' the aliases even if they got messed up somehow.
    """
    for index, details in existing_indexes.items():
        if alias in details.get("aliases", {}):
            yield index, alias


def perform_create_index(indexable, logger):
    """
    Create a new index in ElasticSearch from an indexable instance
    """
    indices_client = IndicesClient(client=ES_CLIENT)
    # Create a new index name, suffixing its name with a timestamp
    new_index = "{:s}_{:s}".format(
        indexable.index_name, timezone.now().strftime("%Y-%m-%d-%Hh%Mm%S.%fs")
    )

    # Create the new index
    logger.info('Creating a new Elasticsearch index "{:s}"...'.format(new_index))
    indices_client.create(index=new_index)

    # The index needs to be closed before we set an analyzer
    indices_client.close(index=new_index)
    indices_client.put_settings(body=ANALYSIS_SETTINGS, index=new_index)
    indices_client.open(index=new_index)

    indices_client.put_mapping(
        body=indexable.mapping, doc_type=indexable.document_type, index=new_index
    )

    # Populate the new index with data provided from our indexable class
    richie_bulk(indexable.get_es_documents(new_index))

    # Return the name of the index we just created in ElasticSearch
    return new_index


def regenerate_indexes(logger):
    """
    Create new indexes for our indexables and replace possible existing indexes with
    a new one only once it has successfully built it.
    """
    # Prepare the client we'll be using to handle indexes
    indices_client = IndicesClient(client=ES_CLIENT)

    # Get all existing indexes once; we'll look up into this list many times
    try:
        existing_indexes = indices_client.get_alias("*")
    except NotFoundError:
        # Provide a fallback empty list so we don't have to check for its existence later on
        existing_indexes = []

    # Create a new index for each of those modules
    # NB: we're mapping perform_create_index which produces side-effects
    indexes_to_create = zip(
        list(map(lambda ix: perform_create_index(ix, logger), ES_INDICES)), ES_INDICES
    )

    # Prepare to alias them so they can be swapped-in for the previous versions
    actions_to_create_aliases = [
        {"add": {"index": index, "alias": ix.index_name}}
        for index, ix in indexes_to_create
    ]

    # Get the previous indexes for every alias
    indexes_to_unalias = reduce(
        lambda acc, ix: acc
        + list(get_indexes_by_alias(existing_indexes, ix.index_name)),
        ES_INDICES,
        [],
    )

    # Prepare to un-alias them so they can be swapped-out for the new versions
    # NB: use chain to flatten the list of generators
    actions_to_delete_aliases = [
        {"remove": {"index": index, "alias": alias}}
        for index, alias in indexes_to_unalias
    ]

    # Identify orphaned indexes
    # NB: we *must* do this before the update_aliases call so we don't immediately prune
    # version n-1 of all our indexes
    useless_indexes = [
        index for index, details in existing_indexes.items() if "aliases" not in details
    ]

    # Replace the old indexes with the new ones in 1 atomic operation to avoid outage
    indices_client.update_aliases(
        dict(actions=actions_to_create_aliases + actions_to_delete_aliases)
    )

    # Cleanup step: do prune older indexes that are now useless
    for useless_index in useless_indexes:
        # Disable keyword arguments checking as elasticsearch-py uses a decorator to list
        # valid query parameters and inject them as kwargs. I won't say a word about this
        # anti-pattern.
        #
        # pylint: disable=unexpected-keyword-arg
        indices_client.delete(index=useless_index, ignore=[400, 404])


def store_es_scripts(logger):
    """
    Iterate over the indexers listed in the settings, import them, and store the scripts
    they define on their "scripts" key in ElasticSearch
    """
    for indexer in ES_INDICES:
        for script_id, script_body in indexer.scripts.items():
            if logger:
                logger.info('Storing script "{:s}"...'.format(script_id))
            ES_CLIENT.put_script(id=script_id, body=script_body)



"""
Declare and configure the models for the courses application
"""
from django.contrib.auth.models import Group
from django.db import models
from django.utils.translation import ugettext_lazy as _

from cms.api import Page
from filer.models import Folder

from ..defaults import ROLE_CHOICES


class PageRole(models.Model):
    """A model to define and control by roles the permissions related to a page."""

    role = models.CharField(
        choices=ROLE_CHOICES,
        max_length=20,
        verbose_name=_("role"),
        help_text=_(
            "A role describes all the permissions that should be "
            "granted to the user group."
        ),
    )
    page = models.ForeignKey(
        to=Page,
        related_name="roles",
        verbose_name=_("page"),
        help_text=_("Page to which this role grants permissions."),
        on_delete=models.CASCADE,
        limit_choices_to={
            # permissions work with draft instances
            "publisher_is_draft": True
        },
    )
    group = models.OneToOneField(
        to=Group,
        related_name="role",
        verbose_name=_("group"),
        help_text=_("User group that this role controls."),
        on_delete=models.PROTECT,
        blank=True,
        editable=False,
    )
    folder = models.OneToOneField(
        to=Folder,
        related_name="role",
        verbose_name=_("filer folder"),
        help_text=_("Filer folder that this role controls."),
        on_delete=models.PROTECT,
        blank=True,
        editable=False,
    )

    class Meta:
        db_table = "richie_page_role"
        verbose_name = _("page role")
        unique_together = ("page", "role")

    def __str__(self):
        """Human representation of a page role."""
        return _("{:s} | {:s}").format(self.get_role_display(), self.page.get_title())

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        """
        Validate the object before saving it and create a group if it does not exist yet.
        """
        self.full_clean(exclude=["group"])
        page_id = str(self.page.id)

        # Create the related group the first time the instance is saved
        if not self.group_id:
            name = str(self)[: Group._meta.get_field("name").max_length]

            if Group.objects.filter(name=name).exists():
                name = f"{name:s} [{page_id:s}]"[
                    : Group._meta.get_field("name").max_length
                ]

            self.group = Group.objects.create(name=name)

        # Create the related filer folder the first time the instance is saved.
        # Why create this folder at the root and not below a parent `organization` folder?
        # - the only way to refer to an existing parent folder is by its name... but it could be
        #   changed by a user via the interface and break the functionality,
        # - the filer search functionality only finds folders at the root, not nested folders.
        if not self.folder_id:
            name = str(self)[: Folder._meta.get_field("name").max_length]

            if Folder.objects.filter(name=name).exists():
                name = f"{name:s} [{page_id:s}]"[
                    : Folder._meta.get_field("name").max_length
                ]

            self.folder = Folder.objects.create(name=name)

        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )



"""
ElasticSearch category document management utilities
"""
from collections import defaultdict

from django.conf import settings
from django.utils import translation

from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.simple_picture.helpers import get_picture_info
from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import Category
from ..forms import ItemSearchForm
from ..text_indexing import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion


class CategoriesIndexer:
    """
    Makes available the parameters the indexer requires as well as functions to shape
    objects getting into and out of ElasticSearch
    """

    document_type = "category"
    index_name = "richie_categories"
    form = ItemSearchForm
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            # Searchable
            # description & title are handled by `MULTILINGUAL_TEXT`
            **{
                "complete.{:s}".format(lang): {
                    "type": "completion",
                    "analyzer": "simple_diacritics_insensitive",
                    # Allow filtering autocomplete results with a kindkey to match the value of
                    # the `kind` field of the object
                    "contexts": [{"name": "kind", "type": "category", "path": "kind"}],
                }
                for lang, _ in settings.LANGUAGES
            },
            "is_meta": {"type": "boolean"},
            "nb_children": {"type": "integer"},
            "path": {"type": "keyword"},
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "logo": {"type": "object", "enabled": False},
            "icon": {"type": "object", "enabled": False},
        },
    }
    scripts = {}
    display_fields = [
        "absolute_url",
        "icon",
        "is_meta",
        "logo",
        "nb_children",
        "path",
        "title.*",
    ]

    @staticmethod
    def get_es_id(page):
        """
        An ID built with the node path and the position of this category in the taxonomy:
            - P: parent, the category has children
            - L: leaf, the category has no children

        For example: a parent category `P_00010002` and its child `L_000100020001`.
        """
        return "{type:s}-{path:s}".format(
            type="L" if page.node.is_leaf() else "P",  # Leaf or Parent?
            path=page.node.path,
        )

    @classmethod
    def get_es_document_for_category(cls, category, index=None, action="index"):
        """Build an Elasticsearch document from the category instance."""
        index = index or cls.index_name

        # Prepare published titles
        titles = {
            t.language: t.title
            for t in Title.objects.filter(page=category.extended_object, published=True)
        }

        # Prepare logo images
        logo_images = {}
        for logo in Picture.objects.filter(
            cmsplugin_ptr__placeholder__page=category.extended_object,
            cmsplugin_ptr__placeholder__slot="logo",
        ):
            language = logo.cmsplugin_ptr.language
            with translation.override(language):
                logo_images[language] = get_picture_info(logo, "logo")

        # Prepare icon images
        icon_images = {}
        for icon in Picture.objects.filter(
            cmsplugin_ptr__placeholder__page=category.extended_object,
            cmsplugin_ptr__placeholder__slot="icon",
        ):
            language = icon.cmsplugin_ptr.language
            with translation.override(language):
                icon_images[language] = get_picture_info(icon, "icon")

        # Prepare description texts
        description = defaultdict(list)
        for simple_text in SimpleText.objects.filter(
            cmsplugin_ptr__placeholder__page=category.extended_object,
            cmsplugin_ptr__placeholder__slot="description",
        ):
            description[simple_text.cmsplugin_ptr.language].append(simple_text.body)

        # Shorcut to the category's page node
        node = category.extended_object.node

        # Find the meta category the current category falls under
        try:
            kind = category.get_meta_category().extended_object.reverse_id
        except Category.DoesNotExist:
            # Meta categories do not have a meta category themselves
            kind = None

        return {
            "_id": cls.get_es_id(category.extended_object),
            "_index": index,
            "_op_type": action,
            "_type": cls.document_type,
            "absolute_url": {
                language: category.extended_object.get_absolute_url(language)
                for language in titles.keys()
            },
            "complete": {
                language: slice_string_for_completion(title)
                for language, title in titles.items()
            },
            "description": {l: " ".join(st) for l, st in description.items()},
            "icon": icon_images,
            "is_meta": bool(
                node.parent is None
                or node.parent.cms_pages.filter(category__isnull=True).exists()
            ),
            "kind": kind,
            "logo": logo_images,
            "nb_children": node.numchild,
            "path": node.path,
            "title": titles,
        }

    @classmethod
    def get_es_documents(cls, index=None, action="index"):
        """
        Loop on all the categories in database and format them for the ElasticSearch index
        """
        index = index or cls.index_name

        for category in Category.objects.filter(
            extended_object__publisher_is_draft=False,
            extended_object__title_set__published=True,
        ).distinct():
            yield cls.get_es_document_for_category(category, index=index, action=action)

    @staticmethod
    def format_es_object_for_api(es_category, best_language):
        """
        Format an category stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        source = es_category["_source"]
        return {
            "icon": get_best_field_language(source["icon"], best_language),
            "id": es_category["_id"],
            "is_meta": source["is_meta"],
            "logo": get_best_field_language(source["logo"], best_language),
            "nb_children": source["nb_children"],
            "path": source["path"],
            "title": get_best_field_language(source["title"], best_language),
        }

    @staticmethod
    def format_es_document_for_autocomplete(es_document, language=None):
        """
        Format a document stored in ES into an easy-to-consume record for autocomplete consumers.
        This method differs from the regular one as objects retrieved from query VS complete
        queries can be formatted differently; and consumers of autocomplete do not need
        full objects.
        """
        return {
            "id": es_document["_id"],
            "kind": "categories",
            "title": get_best_field_language(es_document["_source"]["title"], language),
        }



"""
API endpoints to access organizations through ElasticSearch
"""
from django.conf import settings
from django.utils.translation import get_language_from_request

from elasticsearch.exceptions import NotFoundError
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from .. import ES_CLIENT
from ..defaults import ES_PAGE_SIZE
from ..indexers import ES_INDICES
from ..utils.viewsets import AutocompleteMixin, ViewSetMetadata


class PersonsViewSet(AutocompleteMixin, ViewSet):
    """
    A simple viewset with GET endpoints to fetch persons
    See API Blueprint for details on consumer use.
    """

    _meta = ViewSetMetadata(indexer=ES_INDICES.persons)

    # pylint: disable=no-self-use,unused-argument
    def list(self, request, version):
        """
        Person search endpoint: pass query params to ElasticSearch so it filters
        persons and returns a list of matching items
        """
        # Instantiate the form to allow validation/cleaning
        params_form = self._meta.indexer.form(data=request.query_params)

        # Return a 400 with error information if the query params are not valid
        if not params_form.is_valid():
            return Response(status=400, data={"errors": params_form.errors})

        limit, offset, query = params_form.build_es_query()

        # pylint: disable=unexpected-keyword-arg
        search_query_response = ES_CLIENT.search(
            _source=getattr(self._meta.indexer, "display_fields", "*"),
            index=self._meta.indexer.index_name,
            doc_type=self._meta.indexer.document_type,
            body=query,
            # Directly pass meta-params through as arguments to the ES client
            from_=offset,
            size=limit or getattr(settings, "RICHIE_ES_PAGE_SIZE", ES_PAGE_SIZE),
        )

        # Format the response in a consumer-friendly way
        # NB: if there are 0 hits the query response is formatted the exact same way, only the
        # .hits.hits array is empty.
        response_object = {
            "meta": {
                "count": len(search_query_response["hits"]["hits"]),
                "offset": offset,
                "total_count": search_query_response["hits"]["total"],
            },
            "objects": [
                self._meta.indexer.format_es_object_for_api(
                    person,
                    # Get the best language to return multilingual fields
                    get_language_from_request(request),
                )
                for person in search_query_response["hits"]["hits"]
            ],
        }
        return Response(response_object)

    # pylint: disable=no-self-use,invalid-name,unused-argument
    def retrieve(self, request, pk, version):
        """
        Return a single person by ID
        """
        # Wrap the ES get in a try/catch so we control the exception we emit — it would
        # raise and end up in a 500 error otherwise
        try:
            query_response = ES_CLIENT.get(
                index=self._meta.indexer.index_name,
                doc_type=self._meta.indexer.document_type,
                id=pk,
            )
        except NotFoundError:
            return Response(status=404)

        # Format a clean person object as a response
        return Response(
            self._meta.indexer.format_es_object_for_api(
                query_response,
                # Get the best language we can return multilingual fields in
                get_language_from_request(request),
            )
        )
